#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
import numpy as np
from sklearn.cluster import KMeans
import sys

from numina.array.display.matplotlib_qt import plt
from numina.array.display.matplotlib_qt import patches as patches
from numina.array.display.matplotlib_qt import set_window_geometry
from numina.array.display.pause_debugplot import pause_debugplot
from emirdrp.instrument.csu_configuration import CsuConfiguration
from emirdrp.processing.wavecal.set_wv_parameters import set_wv_parameters

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NBARS
from numina.array.display.pause_debugplot import DEBUGPLOT_CODES


def display_slitlet_arrangement(fileobj,
                                grism=None,
                                spfilter=None,
                                longslits=False,
                                bbox=None,
                                adjust=None,
                                geometry=None,
                                debugplot=0):
    """Display slitlet arrangment from CSUP keywords in FITS header.

    Parameters
    ----------
    fileobj : file object
        FITS or TXT file object.
    grism : str
        Grism name.
    spfilter : str
        Filter name.
    longslits : bool
        If True, display (pseudo) longslits.
    bbox : tuple of 4 floats
        If not None, values for xmin, xmax, ymin and ymax.
    adjust : bool
        Adjust X range according to minimum and maximum csu_bar_left
        and csu_bar_right (note that this option is overriden by 'bbox')
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the Qt backend geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot

    Returns
    -------
    csu_bar_left : list of floats
        Location (mm) of the left bar for each slitlet.
    csu_bar_right : list of floats
        Location (mm) of the right bar for each slitlet, using the
        same origin employed for csu_bar_left (which is not the
        value stored in the FITS keywords.
    csu_bar_slit_center : list of floats
        Middle point (mm) in between the two bars defining a slitlet.
    csu_bar_slit_width : list of floats
        Slitlet width (mm), computed as the distance between the two
        bars defining the slitlet.

    """

    if fileobj.name[-4:] == ".txt":
        if grism is None:
            raise ValueError("Undefined grism!")
        if spfilter is None:
            raise ValueError("Undefined filter!")
        # define CsuConfiguration object
        csu_config = CsuConfiguration()
        csu_config._csu_bar_left = []
        csu_config._csu_bar_right = []
        csu_config._csu_bar_slit_center = []
        csu_config._csu_bar_slit_width = []

        # since the input filename has been opened with argparse in binary
        # mode, it is necessary to close it and open it in text mode
        fileobj.close()
        # read TXT file
        with open(fileobj.name, mode='rt') as f:
            file_content = f.read().splitlines()
        next_id_bar = 1
        for line in file_content:
            if len(line) > 0:
                if line[0] not in ['#']:
                    line_contents = line.split()
                    id_bar = int(line_contents[0])
                    position = float(line_contents[1])
                    if id_bar == next_id_bar:
                        if id_bar <= EMIR_NBARS:
                            csu_config._csu_bar_left.append(position)
                            next_id_bar = id_bar + EMIR_NBARS
                        else:
                            csu_config._csu_bar_right.append(341.5 - position)
                            next_id_bar = id_bar - EMIR_NBARS + 1
                    else:
                        raise ValueError("Unexpected id_bar:" + str(id_bar))

        # compute slit width and center
        for i in range(EMIR_NBARS):
            csu_config._csu_bar_slit_center.append(
                (csu_config._csu_bar_left[i] + csu_config._csu_bar_right[i])/2
            )
            csu_config._csu_bar_slit_width.append(
                csu_config._csu_bar_right[i] - csu_config._csu_bar_left[i]
            )

    else:
        # read input FITS file
        hdulist = fits.open(fileobj.name)
        image_header = hdulist[0].header
        hdulist.close()

        # additional info from header
        grism = image_header['grism']
        spfilter = image_header['filter']

        # define slitlet arrangement
        csu_config = CsuConfiguration.define_from_fits(fileobj)
        if longslits:
            csu_config.display_pseudo_longslits()
            pause_debugplot(debugplot)

    # determine calibration
    if grism in ["J", "OPEN"] and spfilter == "J":
        wv_parameters = set_wv_parameters("J", "J")
    elif grism in ["H", "OPEN"] and spfilter == "H":
        wv_parameters = set_wv_parameters("H", "H")
    elif grism in ["K", "OPEN"] and spfilter == "Ksp":
        wv_parameters = set_wv_parameters("Ksp", "K")
    elif grism in ["LR", "OPEN"] and spfilter == "YJ":
        wv_parameters = set_wv_parameters("YJ", "LR")
    elif grism in ["LR", "OPEN"] and spfilter == "HK":
        wv_parameters = set_wv_parameters("HK", "LR")
    else:
        raise ValueError("Invalid grism + filter configuration")

    crval1 = wv_parameters['poly_crval1_linear']
    cdelt1 = wv_parameters['poly_cdelt1_linear']

    wvmin_useful = wv_parameters['wvmin_useful']
    wvmax_useful = wv_parameters['wvmax_useful']

    # display arrangement
    if abs(debugplot) >= 10:
        print("slit     left    right   center   width   min.wave   max.wave")
        print("====  =======  =======  =======   =====   ========   ========")
        for i in range(EMIR_NBARS):
            ibar = i + 1
            csu_crval1 = crval1(csu_config.csu_bar_slit_center(ibar))
            csu_cdelt1 = cdelt1(csu_config.csu_bar_slit_center(ibar))
            csu_crvaln = csu_crval1 + (EMIR_NAXIS1 - 1) * csu_cdelt1
            if wvmin_useful is not None:
                csu_crval1 = np.amax([csu_crval1, wvmin_useful])
            if wvmax_useful is not None:
                csu_crvaln = np.amin([csu_crvaln, wvmax_useful])
            print("{0:4d} {1:8.3f} {2:8.3f} {3:8.3f} {4:7.3f}   "
                  "{5:8.2f}   {6:8.2f}".format(
                ibar, csu_config.csu_bar_left(ibar),
                csu_config.csu_bar_right(ibar),
                csu_config.csu_bar_slit_center(ibar),
                csu_config.csu_bar_slit_width(ibar),
                csu_crval1, csu_crvaln)
            )
        print(
            "---> {0:8.3f} {1:8.3f} {2:8.3f} {3:7.3f} <- mean (all)".format(
                np.mean(csu_config._csu_bar_left),
                np.mean(csu_config._csu_bar_right),
                np.mean(csu_config._csu_bar_slit_center),
                np.mean(csu_config._csu_bar_slit_width)
            )
        )
        print(
            "---> {0:8.3f} {1:8.3f} {2:8.3f} {3:7.3f} <- mean (odd)".format(
                np.mean(csu_config._csu_bar_left[::2]),
                np.mean(csu_config._csu_bar_right[::2]),
                np.mean(csu_config._csu_bar_slit_center[::2]),
                np.mean(csu_config._csu_bar_slit_width[::2])
            )
        )
        print(
            "---> {0:8.3f} {1:8.3f} {2:8.3f} {3:7.3f} <- mean (even)".format(
                np.mean(csu_config._csu_bar_left[1::2]),
                np.mean(csu_config._csu_bar_right[1::2]),
                np.mean(csu_config._csu_bar_slit_center[1::2]),
                np.mean(csu_config._csu_bar_slit_width[1::2])
            )
        )

    # display slit arrangement
    if abs(debugplot) % 10 != 0:
        fig = plt.figure()
        set_window_geometry(geometry)
        ax = fig.add_subplot(111)
        if bbox is None:
            if adjust:
                xmin = min(csu_config._csu_bar_left)
                xmax = max(csu_config._csu_bar_right)
                dx = xmax - xmin
                if dx == 0:
                    dx = 1
                xmin -= dx/20
                xmax += dx/20
                ax.set_xlim(xmin, xmax)
            else:
                ax.set_xlim(0., 341.5)
            ax.set_ylim(0, 56)
        else:
            ax.set_xlim(bbox[0], bbox[1])
            ax.set_ylim(bbox[2], bbox[3])
        ax.set_xlabel('csu_bar_position (mm)')
        ax.set_ylabel('slit number')
        for i in range(EMIR_NBARS):
            ibar = i + 1
            ax.add_patch(patches.Rectangle(
                (csu_config.csu_bar_left(ibar), ibar-0.5),
                csu_config.csu_bar_slit_width(ibar), 1.0))
            ax.plot([0., csu_config.csu_bar_left(ibar)], [ibar, ibar],
                    '-', color='gray')
            ax.plot([csu_config.csu_bar_right(ibar), 341.5],
                    [ibar, ibar], '-', color='gray')
        plt.title("File: " + fileobj.name + "\ngrism=" + grism +
                  ", filter=" + spfilter)
        pause_debugplot(debugplot, pltshow=True)

    # return results
    return csu_config._csu_bar_left, csu_config._csu_bar_right, \
           csu_config._csu_bar_slit_center, csu_config._csu_bar_slit_width


def display_slitlet_histogram(csu_bar_slit_width,
                              n_clusters=2,
                              geometry=None,
                              debugplot=0):
    """

    Find separations between groups of slitlet widths.

    Parameters
    ----------
    csu_bar_slit_width : numpy array
        Array containing the csu_bar_slit_center values.
    n_clusters : int
        Number of slitlet groups to be sought.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the Qt backend geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot

    Returns
    -------
    TBD

    """

    # protections
    if n_clusters < 2:
        raise ValueError("n_clusters must be >= 2")

    # determine useful slitlets from their widths
    km = KMeans(n_clusters=n_clusters)
    fit = km.fit(np.array(csu_bar_slit_width).reshape(-1, 1))
    separator_list = []
    for i in range(n_clusters - 1):
        peak1 = fit.cluster_centers_[i][0]
        peak2 = fit.cluster_centers_[i+1][0]
        separator = (peak1 + peak2) / 2
        separator_list.append(separator)
        array_csu_bar_slit_width = np.array(csu_bar_slit_width)
        list_ok = []
        list_not_ok = []
        for k in range(EMIR_NBARS):
            if array_csu_bar_slit_width[k] > separator:
                list_ok.append(k + 1)
            else:
                list_not_ok.append(k + 1)
        print('\nNumber of slitlets with width > separator: {}'.format(
            sum(np.array(csu_bar_slit_width) > separator)
        ))
        print(list_ok)
        print('\nNumber of slitlets with width < separator: {}'.format(
            sum(np.array(csu_bar_slit_width) < separator)
        ))
        print(list_not_ok)
        print('\n--->  separator: {0:7.3f}'.format(separator))

    # display histogram
    if abs(debugplot) % 10 != 0:
        fig = plt.figure()
        set_window_geometry(geometry)
        ax = fig.add_subplot(111)
        ax.hist(csu_bar_slit_width, bins=100)
        ax.set_xlabel('slit width (mm)')
        ax.set_ylabel('number of slitlets')
        for separator in separator_list:
            ax.axvline(separator, color='C1', linestyle='dashed')
        pause_debugplot(debugplot, pltshow=True)


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: display arrangement of EMIR CSU bars'
    )

    # positional arguments
    parser.add_argument("filename",
                        help="FITS files (wildcards accepted) or single TXT "
                             "file with CSU configuration from OSP",
                        type=argparse.FileType('rb'),
                        nargs='+')

    # optional arguments
    parser.add_argument("--grism",
                        help="Grism (J, H, K, LR)",
                        choices=["J", "H", "K", "LR"])
    parser.add_argument("--filter",
                        help="Filter (J, H, Ksp, YJ, HK)",
                        choices=["J", "H", "Ksp", "YJ", "HK"])
    parser.add_argument("--n_clusters",
                        help="Display histogram of slitlet widths",
                        default=0, type=int)
    parser.add_argument("--longslits",
                        help="Display (pseudo) longslits built with "
                             "consecutive slitlets",
                        action="store_true")
    parser.add_argument("--bbox",
                        help="Bounding box tuple xmin,xmax,ymin,ymax "
                             "indicating plot limits")
    parser.add_argument("--adjust",
                        help="Adjust X range according to minimum and maximum"
                             " csu_bar_left and csu_bar_right (note that this "
                             "option is overriden by --bbox",
                        action='store_true')
    parser.add_argument("--geometry",
                        help="Tuple x,y,dx,dy indicating window geometry",
                        default="0,0,640,480")
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting & debugging options"
                             " (default=12)",
                        default=12, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # geometry
    if args.geometry is None:
        geometry = None
    else:
        tmp_str = args.geometry.split(",")
        x_geom = int(tmp_str[0])
        y_geom = int(tmp_str[1])
        dx_geom = int(tmp_str[2])
        dy_geom = int(tmp_str[3])
        geometry = x_geom, y_geom, dx_geom, dy_geom

    # read bounding box
    if args.bbox is None:
        bbox = None
    else:
        str_bbox = args.bbox.split(",")
        xmin, xmax, ymin, ymax = [int(str_bbox[i]) for i in range(4)]
        bbox = xmin, xmax, ymin, ymax

    # if input file is a single txt file, assume it is a list of FITS files
    if len(args.filename) == 1:
        list_fits_file_objects = [args.filename[0]]
    else:
        list_fits_file_objects = args.filename

    # total number of files to be examined
    nfiles = len(list_fits_file_objects)

    # declare arrays to store CSU values
    csu_bar_left = np.zeros((nfiles, EMIR_NBARS))
    csu_bar_right = np.zeros((nfiles, EMIR_NBARS))
    csu_bar_slit_center = np.zeros((nfiles, EMIR_NBARS))
    csu_bar_slit_width = np.zeros((nfiles, EMIR_NBARS))

    # display CSU bar arrangement
    for ifile, fileobj in enumerate(list_fits_file_objects):
        print("\nFile " + str(ifile+1) + "/" + str(nfiles) + ": " +
              fileobj.name)
        csu_bar_left[ifile, :], csu_bar_right[ifile, :], \
        csu_bar_slit_center[ifile, :], csu_bar_slit_width[ifile, :] = \
            display_slitlet_arrangement(
                fileobj,
                grism=args.grism,
                spfilter=args.filter,
                longslits=args.longslits,
                bbox=bbox,
                adjust=args.adjust,
                geometry=geometry,
                debugplot=args.debugplot
            )
        if args.n_clusters >= 2:
            display_slitlet_histogram(
                csu_bar_slit_width[ifile, :],
                n_clusters=args.n_clusters,
                geometry=geometry,
                debugplot=args.debugplot
            )

    # print summary of comparison between files
    if nfiles > 1:
        std_csu_bar_left = np.zeros(EMIR_NBARS)
        std_csu_bar_right = np.zeros(EMIR_NBARS)
        std_csu_bar_slit_center = np.zeros(EMIR_NBARS)
        std_csu_bar_slit_width = np.zeros(EMIR_NBARS)
        if args.debugplot >= 10:
            print("\n   STANDARD DEVIATION BETWEEN IMAGES")
            print("slit     left    right   center   width")
            print("====  =======  =======  =======   =====")
            for i in range(EMIR_NBARS):
                ibar = i + 1
                std_csu_bar_left[i] = np.std(csu_bar_left[:, i])
                std_csu_bar_right[i] = np.std(csu_bar_right[:, i])
                std_csu_bar_slit_center[i] = np.std(csu_bar_slit_center[:, i])
                std_csu_bar_slit_width[i] = np.std(csu_bar_slit_width[:, i])
                print("{0:4d} {1:8.3f} {2:8.3f} {3:8.3f} {4:7.3f}".format(
                    ibar,
                    std_csu_bar_left[i],
                    std_csu_bar_right[i],
                    std_csu_bar_slit_center[i],
                    std_csu_bar_slit_width[i]))
            print("====  =======  =======  =======   =====")
            print("MIN: {0:8.3f} {1:8.3f} {2:8.3f} {3:7.3f}".format(
                std_csu_bar_left.min(),
                std_csu_bar_right.min(),
                std_csu_bar_slit_center.min(),
                std_csu_bar_slit_width.min()))
            print("MAX: {0:8.3f} {1:8.3f} {2:8.3f} {3:7.3f}".format(
                std_csu_bar_left.max(),
                std_csu_bar_right.max(),
                std_csu_bar_slit_center.max(),
                std_csu_bar_slit_width.max()))
            print("====  =======  =======  =======   =====")
            print("Total number of files examined:", nfiles)

    # stop program execution
    if len(list_fits_file_objects) > 1:
        pause_debugplot(12, optional_prompt="Press RETURN to STOP")


if __name__ == "__main__":
    main()



#
# Copyright 2013-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Test recipe to find point sources"""

from __future__ import division


import sep
import numpy
import astropy.io.fits as fits
from astropy.convolution import Gaussian2DKernel
from astropy.stats import gaussian_fwhm_to_sigma
from numina.array.fwhm import compute_fwhm_2d_simple
from numina.core import Result, Parameter
from numina.exceptions import RecipeError
import numina.types.array as tarray
from numina.processing.combine import basic_processing_with_combination
from scipy.spatial import KDTree

import emirdrp.datamodel as datamodel
from emirdrp.core.recipe import EmirRecipe
import emirdrp.products as prods
import emirdrp.requirements as reqs

from .procedures import image_box2d


class TestPointSourceRecipe(EmirRecipe):

    # Recipe Requirements
    #
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = reqs.MasterSkyRequirement()

    shift_coordinates = Parameter(True, 'Use header information to'
                                  ' shift the pinhole positions from (0,0) '
                                  'to X_DTU, Y_DTU')
    box_half_size = Parameter(4, 'Half of the computation box size in pixels')
    recenter = Parameter(True, 'Recenter the pinhole coordinates')
    max_recenter_radius = Parameter(2.0, 'Maximum distance for recentering')

    # Recipe Results
    frame = Result(prods.ProcessedImage)
    positions = Result(tarray.ArrayType)
    positions_alt = Result(tarray.ArrayType)
    DTU = Result(tarray.ArrayType)
    filter = Result(str)
    readmode = Result(str)
    ROTANG = Result(float)
    DETPA = Result(float)
    DTUPA = Result(float)
    param_recenter = Result(bool)
    param_max_recenter_radius = Result(float)
    param_box_half_size = Result(float)

    def run(self, rinput):
        self.logger.info('starting processing for object detection')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow=flow)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        self.logger.debug('finding point sources')

        try:
            filtername = hdr['FILTER']
            readmode = hdr['READMODE']
            rotang = hdr['ROTANG']
            detpa = hdr['DETPA']
            dtupa = hdr['DTUPA']
            dtub, dtur = datamodel.get_dtur_from_img(hdulist)
        except KeyError as error:
            self.logger.error(error)
            raise RecipeError(error)

        data = hdulist[0].data

        # Copy needed in numpy 1.7
        # This seems already bitswapped??
        # FIXME: check this works offline/online
        # ndata = data.byteswap().newbyteorder()
        # data = data.byteswap(inplace=True).newbyteorder()

        snr_detect = 5.0
        fwhm = 4.0
        npixels = 15
        box_shape = [64, 64]
        self.logger.info('point source detection2')
        self.logger.info('using internal mask to remove corners')
        # Corners
        mask = numpy.zeros_like(data, dtype='int32')
        mask[2000:, 0:80] = 1
        mask[2028:, 2000:] = 1
        mask[:50, 1950:] = 1
        mask[:100, :50] = 1
        # Remove corner regions

        self.logger.info('compute background map, %s', box_shape)
        bkg = sep.Background(data)

        self.logger.info('reference fwhm is %5.1f pixels', fwhm)
        self.logger.info('detect threshold, %3.1f over background', snr_detect)
        self.logger.info('convolve with gaussian kernel, FWHM %3.1f pixels', fwhm)
        sigma = fwhm * gaussian_fwhm_to_sigma
        #
        kernel = Gaussian2DKernel(sigma)
        kernel.normalize()

        thresh = snr_detect * bkg.globalrms
        data_s = data - bkg.back()
        objects, segmap = sep.extract(data - bkg.back(), thresh, minarea=npixels,
                                      filter_kernel=kernel.array, segmentation_map=True,
                                      mask=mask)
        fits.writeto('segmap.fits', segmap)
        self.logger.info('detected %d objects', len(objects))

        # Hardcoded values
        rs2 = 15.0
        fit_rad = 10.0
        flux_min = 1000.0
        flux_max = 30000.0
        self.logger.debug('Flux limit is %6.1f %6.1f', flux_min, flux_max)
        # FIXME: this should be a view, not a copy
        xall = objects['x']
        yall = objects['y']
        mm = numpy.array([xall, yall]).T
        self.logger.info('computing FWHM')
        # Find objects with pairs inside fit_rad
        kdtree = KDTree(mm)
        nearobjs = (kdtree.query_ball_tree(kdtree, r=fit_rad))
        positions = []
        for idx, obj in enumerate(objects):
            x0 = obj['x']
            y0 = obj['y']
            sl = image_box2d(x0, y0, data.shape, (fit_rad, fit_rad))
            # sl_sky = image_box2d(x0, y0, data.shape, (rs2, rs2))
            part_s = data_s[sl]
            # Logical coordinates
            xx0 = x0 - sl[1].start
            yy0 = y0 - sl[0].start

            _, fwhm_x, fwhm_y = compute_fwhm_2d_simple(part_s, xx0, yy0)

            if min(fwhm_x, fwhm_x) < 3:
                continue
            if flux_min > obj['peak'] or flux_max < obj['peak']:
                continue
            # nobjs is the number of object inside fit_rad
            nobjs = len(nearobjs[idx])

            flag = 0 if nobjs == 1 else 1

            positions.append([idx, x0, y0, obj['peak'], fwhm_x, fwhm_y, flag])

        self.logger.info('saving photometry')
        positions = numpy.array(positions)
        positions_alt = positions
        self.logger.info('end processing for object detection')

        result = self.create_result(frame=hdulist,
                                positions=positions_alt,
                                positions_alt=positions_alt,
                                filter=filtername,
                                DTU=dtub,
                                readmode=readmode,
                                ROTANG=rotang,
                                DETPA=detpa,
                                DTUPA=dtupa,
                                param_recenter=rinput.recenter,
                                param_max_recenter_radius=rinput.max_recenter_radius,
                                param_box_half_size=rinput.box_half_size
                                )
        return result



#
# Copyright 2013-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


"""AIV Recipes for EMIR"""

from __future__ import division

import logging

import numpy
import six
import numina.exceptions
from numina.core import Requirement, Result, Parameter
import numina.types.array as tarray
from numina.processing.combine import basic_processing_with_combination
from scipy import ndimage
from scipy.ndimage.filters import median_filter
from skimage.feature import canny

import emirdrp.datamodel as datamodel
from emirdrp.core import EMIR_PIXSCALE
from emirdrp.core.recipe import EmirRecipe
import emirdrp.requirements as reqs
import emirdrp.products as prods
from .common import normalize, char_slit
from .common import pinhole_char, pinhole_char2

_logger = logging.getLogger(__name__)


class TestMaskRecipe(EmirRecipe):

    # Recipe Requirements
    #
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = reqs.MasterSkyRequirement()

    pinhole_nominal_positions = Requirement(prods.CoordinateList2DType,
                                            'Nominal positions of the pinholes'
                                            )
    shift_coordinates = Parameter(True, 'Use header information to'
                                  ' shift the pinhole positions from (0,0) '
                                  'to X_DTU, Y_DTU')
    box_half_size = Parameter(4, 'Half of the computation box size in pixels')
    recenter = Parameter(True, 'Recenter the pinhole coordinates')
    max_recenter_radius = Parameter(2.0, 'Maximum distance for recentering')

    median_filter_size = Parameter(5, 'Size of the median box')
    canny_sigma = Parameter(3.0, 'Sigma for the canny algorithm')
    obj_min_size = Parameter(200, 'Minimum size of the slit')
    obj_max_size = Parameter(3000, 'Maximum size of the slit')
    slit_size_ratio = Parameter(4.0, 'Minimum ratio between height and width for slits')

    # Recipe Results
    frame = Result(prods.ProcessedImage)
    positions = Result(tarray.ArrayType)
    positions_alt = Result(tarray.ArrayType)
    slitstable = Result(tarray.ArrayType)
    DTU = Result(tarray.ArrayType)
    filter = Result(str)
    readmode = Result(str)
    ROTANG = Result(float)
    DETPA = Result(float)
    DTUPA = Result(float)
    param_recenter = Result(bool)
    param_max_recenter_radius = Result(float)
    param_box_half_size = Result(float)

    def run(self, rinput):
        _logger.info('starting processing for slit detection')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow=flow)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        _logger.debug('finding pinholes')

        try:
            filtername = hdr['FILTER']
            readmode = hdr['READMODE']
            rotang = hdr['ROTANG']
            detpa = hdr['DETPA']
            dtupa = hdr['DTUPA']
            dtub, dtur = datamodel.get_dtur_from_img(hdulist)
        except KeyError as error:
            _logger.error(error)
            raise numina.exceptions.RecipeError(error)

        if rinput.shift_coordinates:
            xdtur, ydtur, zdtur = dtur
            xfac = xdtur / EMIR_PIXSCALE
            yfac = -ydtur / EMIR_PIXSCALE

            vec = numpy.array([yfac, xfac])
            _logger.info('shift is %s', vec)
            ncenters = rinput.pinhole_nominal_positions + vec
        else:
            _logger.info('using pinhole coordinates as they are')
            ncenters = rinput.pinhole_nominal_positions

        _logger.info('pinhole characterization')
        positions = pinhole_char(
            hdulist[0].data,
            ncenters,
            box=rinput.box_half_size,
            recenter_pinhole=rinput.recenter,
            maxdist=rinput.max_recenter_radius
        )

        _logger.info('alternate pinhole characterization')
        positions_alt = pinhole_char2(
            hdulist[0].data, ncenters,
            recenter_pinhole=rinput.recenter,
            recenter_half_box=rinput.box_half_size,
            recenter_maxdist=rinput.max_recenter_radius
        )

        _logger.debug('finding slits')

        # First, prefilter with median
        median_filter_size = rinput.median_filter_size
        canny_sigma = rinput.canny_sigma
        obj_min_size = rinput.obj_min_size
        obj_max_size = rinput.obj_max_size

        data1 = hdulist[0].data
        _logger.debug('Median filter with box %d', median_filter_size)
        data2 = median_filter(data1, size=median_filter_size)

        # Grey level image
        img_grey = normalize(data2)

        # Find edges with canny
        _logger.debug('Find edges with canny, sigma %d', canny_sigma)
        edges = canny(img_grey, sigma=canny_sigma)

        # Fill edges
        _logger.debug('Fill holes')
        fill_slits = ndimage.binary_fill_holes(edges)

        _logger.debug('Label objects')
        label_objects, nb_labels = ndimage.label(fill_slits)
        _logger.debug('%d objects found', nb_labels)
        # Filter on the area of the labeled region
        # Perhaps we could ignore this filtering and
        # do it later?
        _logger.debug('Filter objects by size')
        # Sizes of regions
        sizes = numpy.bincount(label_objects.ravel())

        _logger.debug('Min size is %d', obj_min_size)
        _logger.debug('Max size is %d', obj_max_size)

        mask_sizes = (sizes > obj_min_size) & (sizes < obj_max_size)

        # Filter out regions
        nids, = numpy.where(mask_sizes)

        mm = numpy.in1d(label_objects, nids)
        mm.shape = label_objects.shape

        fill_slits_clean = numpy.where(mm, 1, 0)

        # and relabel
        _logger.debug('Label filtered objects')
        relabel_objects, nb_labels = ndimage.label(fill_slits_clean)
        _logger.debug('%d objects found after filtering', nb_labels)
        ids = list(six.moves.range(1, nb_labels + 1))

        _logger.debug('Find regions and centers')
        regions = ndimage.find_objects(relabel_objects)
        centers = ndimage.center_of_mass(data2, labels=relabel_objects,
                                         index=ids
                                         )

        table = char_slit(data2, regions,
                          slit_size_ratio=rinput.slit_size_ratio
                          )

        result = self.create_result(frame=hdulist,
                                    positions=positions,
                                    positions_alt=positions_alt,
                                    slitstable=table,
                                    filter=filtername,
                                    DTU=dtub,
                                    readmode=readmode,
                                    ROTANG=rotang,
                                    DETPA=detpa,
                                    DTUPA=dtupa,
                                    param_recenter=rinput.recenter,
                                    param_max_recenter_radius=rinput.max_recenter_radius,
                                    param_box_half_size=rinput.box_half_size
                                    )
        return result




import numpy
import pytest
import astropy.units as U

from ..dtuconf import DtuConf, DtuAxis
from ..dtuconf import managed_ndig
from ..dtuconf import average, apply_on_axis

# header fragment
def dtu_header():
    header = {}
    header['XDTU_F'] = 0.922918
    header['YDTU_F'] = 0.971815
    header['ZDTU_F'] = 1.0
    header['XDTU'] = -205.679000854492
    header['YDTU'] = -24.4878005981445
    header['ZDTU'] = -463.765991210938
    header['XDTU_0'] = -205.651000976562
    header['YDTU_0'] = -24.4794006347656
    header['ZDTU_0'] = -463.821014404297
    return header


def dtu_header2():
    header = {}
    header['XDTU_F'] = 0.922918
    header['YDTU_F'] = 0.971815
    header['ZDTU_F'] = 1.0
    header['XDTU'] = -215.670
    header['YDTU'] = -241.487
    header['ZDTU'] = -263.765
    header['XDTU_0'] = -205.651
    header['YDTU_0'] = -24.4794
    header['ZDTU_0'] = -422.821
    return header


def test_dtuc():

    header = dtu_header()

    dtuconf = DtuConf.from_header(header)

    assert isinstance(dtuconf, DtuConf)


def test_dtuc_shift():
    import emirdrp.instrument.constants as cons

    # Value in microns in DTU coords
    header = dtu_header()
    expected = [-17.206285211909773, -0.7186057740102463, 0.055023193358977096]
    dtuconf = DtuConf.from_header(header)

    coor_r = dtuconf.coor_r
    assert numpy.allclose(coor_r, expected)

    expected_pix = [ 0.03992254, -0.95590473,  0.00305684]
    # Value in pixels in image coords
    trans3 = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]  # T3 = T2 * T1

    vec = numpy.dot(trans3, dtuconf.coor_r) * U.micron / cons.EMIR_PIXSIZE
    assert numpy.allclose(vec, expected_pix)


def test_dtuaxis_raise():
    with pytest.raises(ValueError):
        DtuAxis("R", 200.0)


def test_dtuaxis_header():

    header = dtu_header()

    dtuaxis_x = DtuAxis.from_header(header, name='X')

    assert isinstance(dtuaxis_x, DtuAxis)


def test_dtuaxis_header_raise():

    header = dtu_header()

    with pytest.raises(ValueError):
        DtuAxis.from_header(header, name='R')


def test_dtur():
    header = dtu_header()
    dtuconf = DtuConf.from_header(header)
    x_dtu = [-205.679000854492, -24.4878005981445, -463.765991210938]
    x_dtu_r =  [-17.206285211909773, -0.7186057740102463, 0.055023193358977096]

    assert numpy.allclose(x_dtu, dtuconf.coor)

    assert numpy.allclose(x_dtu_r, dtuconf.coor_r)


def test_dtu_formatter():
    expected = (
        "<DtuConf instance>\n"
        "- XDTU..: -205.679\n"
        "- YDTU..:  -24.488\n"
        "- ZDTU..: -463.766\n"
        "- XDTU_0: -205.651\n"
        "- YDTU_0:  -24.479\n"
        "- ZDTU_0: -463.821\n"
        "- XDTU_F:    0.923\n"
        "- YDTU_F:    0.972\n"
        "- ZDTU_F:    1.000\n"
    )
    header = dtu_header()
    dtuconf = DtuConf.from_header(header)
    val = dtuconf.describe()
    assert val == expected


def test_dtu_managed():
    header = dtu_header()
    dtuconf = DtuConf.from_header(header)
    with managed_ndig(dtuconf, 4):
        assert dtuconf.get_ndig() == 4

    assert dtuconf.get_ndig() == 3


def test_dtu_apply():
    axis1 = DtuAxis("X", 100.0)
    axis2 = DtuAxis("X", 120.0)

    axis3 = apply_on_axis(numpy.mean, [axis1, axis2])
    assert axis3.coor == 110.0
    assert axis3.coor_f == 1.0
    assert axis3.coor_0 == 0.0
    assert axis3.name == "X"


def test_dtu_apply0():
    """Test on an empty list"""
    axis3 = apply_on_axis(numpy.mean, [])
    assert numpy.isnan(axis3.coor)
    assert numpy.isnan(axis3.coor_f)
    assert numpy.isnan(axis3.coor_0)
    assert axis3.name == "X"


def test_dtu_average():
    header1 = dtu_header()
    dtuconf1 = DtuConf.from_header(header1)
    header2 = dtu_header2()
    dtuconf2 = DtuConf.from_header(header2)

    dtuconf = average(dtuconf1, dtuconf2)

    assert isinstance(dtuconf, DtuConf)

def test_dtu_average0():

    dtuconf = average()

    assert isinstance(dtuconf, DtuConf)



#
# Copyright 2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Useful X-axis pixels removing +/- npixaround pixels around each OH line"""

import numpy as np
import pkgutil
from six import StringIO

from numina.array.display.ximshow import ximshow
from numina.array.display.pause_debugplot import pause_debugplot

from emirdrp.processing.wavecal.get_islitlet import get_islitlet


def useful_mos_xpixels(reduced_mos_data,
                       base_header,
                       vpix_region,
                       npix_removed_near_ohlines=0,
                       list_valid_wvregions=None,
                       debugplot=0):
    """Useful X-axis pixels removing +/- npixaround pixels around each OH line
    """

    # get wavelength calibration from image header
    naxis1 = base_header['naxis1']
    naxis2 = base_header['naxis2']
    crpix1 = base_header['crpix1']
    crval1 = base_header['crval1']
    cdelt1 = base_header['cdelt1']

    # check vertical region
    nsmin = int(vpix_region[0] + 0.5)
    nsmax = int(vpix_region[1] + 0.5)
    if nsmin > nsmax:
        raise ValueError('vpix_region values in wrong order')
    elif nsmin < 1 or nsmax > naxis2:
        raise ValueError('vpix_region outside valid range')

    # minimum and maximum pixels in the wavelength direction
    islitlet_min = get_islitlet(nsmin)
    ncmin = base_header['jmnslt{:02d}'.format(islitlet_min)]
    ncmax = base_header['jmxslt{:02d}'.format(islitlet_min)]
    islitlet_max = get_islitlet(nsmax)
    if islitlet_max > islitlet_min:
        for islitlet in range(islitlet_min + 1, islitlet_max + 1):
            ncmin_ = base_header['jmnslt{:02d}'.format(islitlet)]
            ncmax_ = base_header['jmnslt{:02d}'.format(islitlet)]
            ncmin = min(ncmin, ncmin_)
            ncmax = max(ncmax, ncmax_)

    # pixels within valid regions
    xisok_wvreg = np.zeros(naxis1, dtype='bool')
    if list_valid_wvregions is None:
        for ipix in range(ncmin, ncmax + 1):
            xisok_wvreg[ipix - 1] = True
    else:
        for wvregion in list_valid_wvregions:
            wvmin = float(wvregion[0])
            wvmax = float(wvregion[1])
            if wvmin > wvmax:
                raise ValueError('wvregion values in wrong order:'
                                 ' {}, {}'.format(wvmin, wvmax))
            minpix = int((wvmin - crval1) / cdelt1 + crpix1 + 0.5)
            maxpix = int((wvmax - crval1) / cdelt1 + crpix1 + 0.5)
            for ipix in range(minpix, maxpix + 1):
                if 1 <= ipix <= naxis1:
                    xisok_wvreg[ipix - 1] = True
        if np.sum(xisok_wvreg) < 1:
            raise ValueError('no valid wavelength ranges provided')

    # pixels affected by OH lines
    xisok_oh = np.ones(naxis1, dtype='bool')
    if int(npix_removed_near_ohlines) > 0:
        dumdata = pkgutil.get_data(
            'emirdrp.instrument.configs',
            'Oliva_etal_2013.dat'
        )
        oh_lines_tmpfile = StringIO(dumdata.decode('utf8'))
        catlines = np.genfromtxt(oh_lines_tmpfile)
        catlines_all_wave = np.concatenate(
            (catlines[:, 1], catlines[:, 0]))
        for waveline in catlines_all_wave:
            expected_pixel = int(
                (waveline - crval1) / cdelt1 + crpix1 + 0.5)
            minpix = expected_pixel - int(npix_removed_near_ohlines)
            maxpix = expected_pixel + int(npix_removed_near_ohlines)
            for ipix in range(minpix, maxpix + 1):
                if 1 <= ipix <= naxis1:
                    xisok_oh[ipix - 1] = False

    # pixels in valid regions not affected by OH lines
    xisok = np.logical_and(xisok_wvreg, xisok_oh)
    naxis1_effective = np.sum(xisok)
    if naxis1_effective < 1:
        raise ValueError('no valid wavelength range available after '
                         'removing OH lines')

    if abs(debugplot) in [21, 22]:
        slitlet2d = reduced_mos_data[(nsmin - 1):nsmax, :].copy()
        ximshow(slitlet2d,
                title='Rectified region',
                first_pixel=(1, nsmin),
                crval1=crval1, cdelt1=cdelt1, debugplot=debugplot)
        ax = ximshow(slitlet2d,
                     title='Rectified region\nafter blocking '
                           'removed wavelength ranges',
                     first_pixel=(1, nsmin),
                     crval1=crval1, cdelt1=cdelt1, show=False)
        for idum in range(1, naxis1 + 1):
            if not xisok_wvreg[idum - 1]:
                ax.plot([idum, idum], [nsmin, nsmax], 'g-')
        pause_debugplot(debugplot, pltshow=True)
        ax = ximshow(slitlet2d,
                     title='Rectified slitlet\nuseful regions after '
                           'removing OH lines',
                     first_pixel=(1, nsmin),
                     crval1=crval1, cdelt1=cdelt1, show=False)
        for idum in range(1, naxis1 + 1):
            if not xisok[idum - 1]:
                ax.plot([idum, idum], [nsmin, nsmax], 'm-')
        pause_debugplot(debugplot, pltshow=True)

    return xisok



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#


import numpy
from numina.array.utils import coor_to_pix_1d
from scipy.ndimage import convolve1d
from scipy.ndimage.filters import median_filter

import emirdrp.instrument.distortions as dist
from emirdrp.core import EMIR_PIXSCALE, EMIR_NBARS, EMIR_RON
from emirdrp.processing.bardetect import char_bar_peak_l, char_bar_peak_r


def median_filtering(hdulist, mfilter_size):
    # Processed array
    arr = hdulist[0].data
    arr_median = median_filter(arr, size=(1, mfilter_size))
    arr_median = median_filter(arr_median, size=(mfilter_size, 1))
    return arr_median


def find_bars(hdulist, bars_nominal_positions, csupos, dtur,
              average_box_row_size=7, average_box_col_size=21, fit_peak_npoints=3,
              median_filter_size=5, logger=None):

    logger.debug('filtering image')
    # Processed array
    arr_median = median_filtering(hdulist, median_filter_size)

    xfac = dtur[0] / EMIR_PIXSCALE
    yfac = -dtur[1] / EMIR_PIXSCALE

    vec = [yfac, xfac]
    logger.debug('DTU shift is %s', vec)

    # and the table of approx positions of the slits
    barstab = bars_nominal_positions
    # Currently, we only use fields 0 and 2
    # of the nominal positions file

    # Number or rows used
    # These other parameters cab be tuned also
    bstart = 1
    bend = 2047
    logger.debug('ignoring columns outside %d - %d', bstart, bend - 1)

    # extract a region to average
    # wy = (average_box_row_size // 2)
    # wx = (average_box_col_size // 2)
    # logger.debug('extraction window is %d rows, %d cols', 2 * wy + 1, 2 * wx + 1)
    # Fit the peak with these points
    # wfit = 2 * (fit_peak_npoints // 2) + 1
    # logger.debug('fit with %d points', wfit)

    # Minimum threshold
    threshold = 5 * EMIR_RON
    # Savitsky and Golay (1964) filter to compute the X derivative
    # scipy >= xx has a savgol_filter function
    # for compatibility we do it manually

    allpos = {}
    slits = numpy.zeros((EMIR_NBARS, 8), dtype='float')

    logger.info('find peaks in derivative image')
    for ks in [3, 5, 7, 9]:
        logger.debug('kernel size is %d', ks)
        # S and G kernel for derivative
        kw = ks * (ks * ks - 1) / 12.0
        coeffs_are = -numpy.arange((1 - ks) // 2, (ks - 1) // 2 + 1) / kw
        logger.debug('kernel weights are %s', coeffs_are)

        logger.debug('derive image in X direction')
        arr_deriv = convolve1d(arr_median, coeffs_are, axis=-1)
        # self.save_intermediate_array(arr_deriv, 'deriv_image_k%d.fits' % ks)
        # Axis 0 is
        #
        # logger.debug('derive image in Y direction (with kernel=3)')
        # arr_deriv_alt = convolve1d(arr_median_alt, ypos3_kernel, axis=0)

        positions = []
        logger.info('using bar parameters')
        for idx in range(EMIR_NBARS):
            params_l = barstab[idx]
            params_r = barstab[idx + EMIR_NBARS]
            lbarid = int(params_l[0])

            # CSUPOS for this bar
            rbarid = lbarid + EMIR_NBARS
            current_csupos_l = csupos[lbarid - 1]
            current_csupos_r = csupos[rbarid - 1]
            logger.debug('CSUPOS for bar %d is %f', lbarid, current_csupos_l)
            logger.debug('CSUPOS for bar %d is %f', rbarid, current_csupos_r)

            ref_y_coor_virt = params_l[1]  # Do I need to add vec[1]?
            ref_x_l_coor_virt = params_l[3] + current_csupos_l * params_l[2]
            ref_x_r_coor_virt = params_r[3] + current_csupos_r * params_r[2]
            # Transform to REAL..
            ref_x_l_coor, ref_y_l_coor = dist.exvp(ref_x_l_coor_virt, ref_y_coor_virt)
            ref_x_r_coor, ref_y_r_coor = dist.exvp(ref_x_r_coor_virt, ref_y_coor_virt)
            # FIXME: check if DTU has to be applied
            # ref_y_coor = ref_y_coor + vec[1]
            prow = coor_to_pix_1d(ref_y_l_coor) - 1
            fits_row = prow + 1  # FITS pixel index

            logger.debug('looking for bars with ids %d - %d', lbarid, rbarid)
            logger.debug('ref Y virtual position is %7.2f', ref_y_coor_virt)
            logger.debug('ref X virtual positions are %7.2f %7.2f', ref_x_l_coor_virt, ref_x_r_coor_virt)
            logger.debug('ref X positions are %7.2f %7.2f', ref_x_l_coor, ref_x_r_coor)
            logger.debug('ref Y positions are %7.2f %7.2f', ref_y_l_coor, ref_y_r_coor)
            # if ref_y_coor is outlimits, skip this bar
            # ref_y_coor is in FITS format
            if (ref_y_l_coor >= 2047 + 16) or (ref_y_l_coor <= 1 - 16):
                logger.debug('reference y position is outlimits, skipping')
                positions.append([lbarid, fits_row, fits_row, fits_row, 1, 1, 0, 3])
                positions.append([rbarid, fits_row, fits_row, fits_row, 1, 1, 0, 3])
                continue

            # minimal width of the slit
            minwidth = 0.9
            if abs(ref_x_l_coor_virt - ref_x_r_coor_virt) < minwidth:
                # FIXME: if width < minwidth fit peak in image
                logger.debug('slit is less than %d virt pixels, skipping', minwidth)
                positions.append([lbarid, fits_row, fits_row, fits_row, 1, 1, 0, 3])
                positions.append([rbarid, fits_row, fits_row, fits_row, 1, 1, 0, 3])
                continue

            # Left bar
            # Dont add +1 to virtual pixels
            logger.debug('measure left border (%d)', lbarid)
            regionw = 10
            bstart1 = coor_to_pix_1d(ref_x_l_coor - regionw)
            bend1 = coor_to_pix_1d(ref_x_l_coor + regionw) + 1
            centery, centery_virt, xpos1, xpos1_virt, fwhm, st = char_bar_peak_l(arr_deriv,
                                                                                 prow, bstart1, bend1,
                                                                                 threshold)

            insert1 = [lbarid, centery + 1, centery_virt, fits_row, xpos1 + 1, xpos1_virt, fwhm, st]
            positions.append(insert1)

            # Right bar
            # Dont add +1 to virtual pixels
            logger.debug('measure rigth border (%d)', rbarid)
            bstart2 = coor_to_pix_1d(ref_x_r_coor - regionw)
            bend2 = coor_to_pix_1d(ref_x_r_coor + regionw) + 1
            centery, centery_virt, xpos2, xpos2_virt, fwhm, st = char_bar_peak_r(arr_deriv, prow, bstart2, bend2,
                                                                                 threshold)
            # This centery/centery_virt should be equal to ref_y_coor_virt
            insert2 = [rbarid, centery + 1, centery_virt, fits_row, xpos2 + 1, xpos2_virt, fwhm, st]
            positions.append(insert2)

            # FIXME: hardcoded value
            y1_virt = ref_y_coor_virt - 16.242
            y2_virt = ref_y_coor_virt + 16.242
            _, y1 = dist.exvp(xpos1_virt + 1, y1_virt + 1)
            _, y2 = dist.exvp(xpos2_virt + 1, y2_virt + 1)

            # Update positions

            msg = 'bar %d, centroid-y %9.4f centroid-y virt %9.4f, ' \
                  'row %d, x-pos %9.4f x-pos virt %9.4f, FWHM %6.3f, status %d'
            logger.debug(msg, *positions[-2])
            logger.debug(msg, *positions[-1])

            if ks == 5:
                slits[lbarid - 1] = numpy.array([xpos1, y2, xpos2, y2, xpos2, y1, xpos1, y1])
                # FITS coordinates
                slits[lbarid - 1] += 1.0
                logger.debug('inserting bars %d-%d into "slits"', lbarid, rbarid)

        allpos[ks] = numpy.asarray(positions, dtype='float')  # GCS doesn't like lists of lists

    return allpos, slits


def slits_to_ds9_reg(ds9reg, slits):
    """Transform fiber traces to ds9-region format.

    Parameters
    ----------
    ds9reg : BinaryIO
        Handle to output file name in ds9-region format.
    """

    # open output file and insert header

    ds9reg.write('# Region file format: DS9 version 4.1\n')
    ds9reg.write(
        'global color=green dashlist=8 3 width=1 font="helvetica 10 '
        'normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 '
        'move=1 delete=1 include=1 source=1\n')
    ds9reg.write('physical\n')

    for idx, slit in enumerate(slits, 1):
        xpos1, y2, xpos2, y2, xpos2, y1, xpos1, y1 = slit
        xc = 0.5 * (xpos1 + xpos2) + 1
        yc = 0.5 * (y1 + y2) + 1
        xd = (xpos2 - xpos1)
        yd = (y2 - y1)
        ds9reg.write('box({0},{1},{2},{3},0)\n'.format(xc, yc, xd, yd))
        ds9reg.write('# text({0},{1}) color=red text={{{2}}}\n'.format(xpos1 - 5, yc, idx))
        ds9reg.write('# text({0},{1}) color=red text={{{2}}}\n'.format(xpos2 + 5, yc, idx + EMIR_NBARS))



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
import numpy as np
import sys

from numina.tools.arg_file_is_new import arg_file_is_new

from emirdrp.instrument.dtu_configuration import DtuConfiguration
from emirdrp.products import RectWaveCoeff
from emirdrp.processing.wavecal.slitlet2d import Slitlet2D
from emirdrp.tools.nscan_minmax_frontiers import nscan_minmax_frontiers

from .rect_wpoly_for_mos import islitlet_progress
from .save_ndarray_to_fits import save_ndarray_to_fits

from emirdrp.core import EMIR_NBARS
from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2
from emirdrp.core import EMIR_NPIXPERSLIT_RECTIFIED

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: apply rectification polynomials '
                    'for the CSU configuration of a particular image'
    )

    # required arguments
    parser.add_argument("fitsfile",
                        help="Input FITS file",
                        type=argparse.FileType('rb'))
    parser.add_argument("--rectwv_coeff", required=True,
                        help="Input JSON file with rectification and "
                             "wavelength calibration coefficients",
                        type=argparse.FileType('rt'))
    parser.add_argument("--outfile", required=True,
                        help="Output FITS file with rectified image",
                        type=lambda x: arg_file_is_new(parser, x, mode='wb'))

    # optional arguments
    parser.add_argument("--resampling",
                        help="Resampling method: 1 -> nearest neighbor, "
                             "2 -> linear interpolation (default)",
                        default=2, type=int,
                        choices=(1, 2))
    parser.add_argument("--ignore_dtu_configuration",
                        help="Ignore DTU configurations differences between "
                             "transformation and input image",
                        action="store_true")
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting & debugging options"
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # read calibration structure from JSON file
    rectwv_coeff = RectWaveCoeff._datatype_load(
        args.rectwv_coeff.name)

    # read FITS image and its corresponding header
    hdulist = fits.open(args.fitsfile)
    header = hdulist[0].header
    image2d = hdulist[0].data
    hdulist.close()

    # protections
    naxis2, naxis1 = image2d.shape
    if naxis1 != header['naxis1'] or naxis2 != header['naxis2']:
        print('>>> NAXIS1:', naxis1)
        print('>>> NAXIS2:', naxis2)
        raise ValueError('Something is wrong with NAXIS1 and/or NAXIS2')
    if abs(args.debugplot) >= 10:
        print('>>> NAXIS1:', naxis1)
        print('>>> NAXIS2:', naxis2)

    # check that the input FITS file grism and filter match
    filter_name = header['filter']
    if filter_name != rectwv_coeff.tags['filter']:
        raise ValueError("Filter name does not match!")
    grism_name = header['grism']
    if grism_name != rectwv_coeff.tags['grism']:
        raise ValueError("Filter name does not match!")
    if abs(args.debugplot) >= 10:
        print('>>> grism.......:', grism_name)
        print('>>> filter......:', filter_name)

    # check that the DTU configurations are compatible
    dtu_conf_fitsfile = DtuConfiguration.define_from_fits(args.fitsfile)
    dtu_conf_jsonfile = DtuConfiguration.define_from_dictionary(
        rectwv_coeff.meta_info['dtu_configuration'])
    if dtu_conf_fitsfile != dtu_conf_jsonfile:
        print('DTU configuration (FITS file):\n\t', dtu_conf_fitsfile)
        print('DTU configuration (JSON file):\n\t', dtu_conf_jsonfile)
        if args.ignore_dtu_configuration:
            print('WARNING: DTU configuration differences found!')
        else:
            raise ValueError("DTU configurations do not match!")
    else:
        if abs(args.debugplot) >= 10:
            print('>>> DTU Configuration match!')
            print(dtu_conf_fitsfile)

    # valid slitlet numbers
    list_valid_islitlets = list(range(1, EMIR_NBARS + 1))
    for idel in rectwv_coeff.missing_slitlets:
        list_valid_islitlets.remove(idel)
    if abs(args.debugplot) >= 10:
        print('>>> valid slitlet numbers:\n', list_valid_islitlets)

    naxis2_enlarged = EMIR_NBARS * EMIR_NPIXPERSLIT_RECTIFIED
    image2d_rectified = np.zeros((naxis2_enlarged, EMIR_NAXIS1))
    image2d_unrectified = np.zeros((EMIR_NAXIS2, EMIR_NAXIS1))

    for islitlet in list(range(1, EMIR_NBARS + 1)):
        if islitlet in list_valid_islitlets:
            if args.debugplot == 0:
                islitlet_progress(islitlet, EMIR_NBARS, ignore=False)

            # define Slitlet2D object
            slt = Slitlet2D(islitlet=islitlet,
                            rectwv_coeff=rectwv_coeff,
                            debugplot=args.debugplot)

            # extract 2D image corresponding to the selected slitlet: note that
            # in this case we are not using select_unrectified_slitlets()
            # because it introduces extra zero pixels in the slitlet frontiers
            slitlet2d = slt.extract_slitlet2d(image2d)

            # rectify image
            slitlet2d_rect = slt.rectify(slitlet2d,
                                         resampling=args.resampling)

            # minimum and maximum useful row in the full 2d rectified image
            # (starting from 0)
            i1 = slt.iminslt - 1
            i2 = slt.imaxslt

            # minimum and maximum scan in the rectified slitlet
            # (in pixels, from 1 to NAXIS2)
            ii1 = slt.min_row_rectified
            ii2 = slt.max_row_rectified + 1

            # save rectified slitlet in its corresponding location within
            # the full 2d rectified image
            image2d_rectified[i1:i2, :] = slitlet2d_rect[ii1:ii2, :]

            # ---

            # unrectify image
            slitlet2d_unrect = slt.rectify(slitlet2d_rect,
                                           resampling=args.resampling,
                                           inverse=True)

            # minimum and maximum useful scan (pixel in the spatial direction)
            # for the rectified slitlet
            nscan_min, nscan_max = nscan_minmax_frontiers(
                slt.y0_frontier_lower,
                slt.y0_frontier_upper,
                resize=False
            )
            ii1 = nscan_min - slt.bb_ns1_orig
            ii2 = nscan_max - slt.bb_ns1_orig + 1

            j1 = slt.bb_nc1_orig - 1
            j2 = slt.bb_nc2_orig
            i1 = slt.bb_ns1_orig - 1 + ii1
            i2 = i1 + ii2 - ii1

            image2d_unrectified[i1:i2, j1:j2] = slitlet2d_unrect[ii1:ii2, :]

        else:
            if args.debugplot == 0:
                islitlet_progress(islitlet, EMIR_NBARS, ignore=True)

    if args.debugplot == 0:
        print('OK!')

    save_ndarray_to_fits(
        array=[image2d_rectified, image2d_unrectified],
        file_name=args.outfile,
        cast_to_float=[True] * 2,
        overwrite=True
    )
    print('>>> Saving file ' + args.outfile.name)


if __name__ == "__main__":
    main()



#
# Copyright 2008-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

from astropy.io import fits
import numpy as np


class DtuConfiguration(object):
    """Detector Translation Unit (DTU) Configuration class definition.

    Attributes
    ----------
    xdtu : float
        XDTU fits keyword value.
    ydtu : float
        YDTU fits keyword value.
    zdtu : float
        ZDTU fits keyword value.
    xdtu_0 : float
        XDTU_0 fits keyword value.
    ydtu_0 : float
        YDTU_0 fits keyword value.
    zdtu_0 : float
        ZDTU_0 fits keyword value.

    """

    def __init__(self):
        self.xdtu = None
        self.ydtu = None
        self.zdtu = None
        self.xdtu_0 = None
        self.ydtu_0 = None
        self.zdtu_0 = None

    def __str__(self):
        # note: set the number of decimal figures in output to the precision
        # value employed in __eq__() function below
        output = "<DtuConfiguration instance>\n"
        strdum = "- XDTU..: {0:8.3f}\n".format(self.xdtu)
        output += strdum
        strdum = "- YDTU..: {0:8.3f}\n".format(self.ydtu)
        output += strdum
        strdum = "- ZDTU..: {0:8.3f}\n".format(self.zdtu)
        output += strdum
        strdum = "- XDTU_0: {0:8.3f}\n".format(self.xdtu_0)
        output += strdum
        strdum = "- YDTU_0: {0:8.3f}\n".format(self.ydtu_0)
        output += strdum
        strdum = "- ZDTU_0: {0:8.3f}".format(self.zdtu_0)
        output += strdum
        return output

    def __eq__(self, other):
        # note: set the precision (number of decimal places) to the same
        # number employed in __str__() function above to print out member
        # values
        if isinstance(other, DtuConfiguration):
            ndig = 3
            result = \
                (round(self.xdtu, ndig) == round(other.xdtu, ndig)) and \
                (round(self.ydtu, ndig) == round(other.ydtu, ndig)) and \
                (round(self.zdtu, ndig) == round(other.zdtu, ndig)) and \
                (round(self.xdtu_0, ndig) == round(other.xdtu_0, ndig)) and \
                (round(self.ydtu_0, ndig) == round(other.ydtu_0, ndig)) and \
                (round(self.zdtu_0, ndig) == round(other.zdtu_0, ndig))
            return result
        return NotImplemented

    def __ne__(self, other):
        result = not self.__eq__(other)
        return result

    @classmethod
    def define_from_fits(cls, fitsobj, extnum=0):
        """Define class object from header information in FITS file.

        Parameters
        ----------
        fitsobj: file object
            FITS file whose header contains the DTU information
            needed to initialise the members of this class.
        extnum : None or int or str
            Extension number or extension name. If None, select 'MECS' extension
            if available. If not, use primary extension

        """

        # read input FITS file
        with fits.open(fitsobj) as hdulist:
            if extnum is None:
                if 'MECS' in hdulist:
                    extnum = 'MECS'
                else:
                    extnum = 0
            image_header = hdulist[extnum].header
            return cls.define_from_header(image_header)

    @classmethod
    def define_from_img(cls, img):
        """Define class object from header information in HDUList object.

        Parameters
        ----------
        img: HDUList
            FITS image whose header contains the DTU information
            needed to initialise the members of this class.

        """

        # read input FITS file
        if 'MECS' in img:
            extnum = 'MECS'
        else:
            extnum = 0
        image_header = img[extnum].header
        return cls.define_from_header(image_header)

    @classmethod
    def define_from_header(cls, image_header):
        """Define class object directly from FITS header.

        Parameters
        ----------
        image_header : instance of hdulist.header
            Header content from a FITS file.

        """

        return cls.define_from_dictionary(image_header)

    @classmethod
    def define_from_dictionary(cls, inputdict):
        """Define class object from dictionary.

        Parameters
        ----------
        inputdict : dictionary like object
            Dictionary like object defining each member of the class.

        """

        self = DtuConfiguration()
        for item in self.__dict__:
            self.__dict__[item] = inputdict[item]
        return self

    @classmethod
    def define_from_values(cls, xdtu, ydtu, zdtu, xdtu_0, ydtu_0, zdtu_0):
        """Define class object from from provided values.

        Parameters
        ----------
        xdtu : float
            XDTU fits keyword value.
        ydtu : float
            YDTU fits keyword value.
        zdtu : float
            ZDTU fits keyword value.
        xdtu_0 : float
            XDTU_0 fits keyword value.
        ydtu_0 : float
            YDTU_0 fits keyword value.
        zdtu_0 : float
            ZDTU_0 fits keyword value.

        """

        self = DtuConfiguration()
        # define DTU variables
        self.xdtu = xdtu
        self.ydtu = ydtu
        self.zdtu = zdtu
        self.xdtu_0 = xdtu_0
        self.ydtu_0 = ydtu_0
        self.zdtu_0 = zdtu_0
        return self

    def closeto(self, other, abserror):
        """Check that all the members are equal within provided absolute error.

        Parameters
        ----------
        other : DtuConfiguration object
            DTU configuration instance to be compared with self.
        abserror : float
            Absolute maximum allowed error.

        Returns
        -------
        result : bool
            True is all members are within the specified maximum
            absolute error

        """

        result = \
            (abs(self.xdtu - other.xdtu) <= abserror) and \
            (abs(self.ydtu - other.ydtu) <= abserror) and \
            (abs(self.zdtu - other.zdtu) <= abserror) and \
            (abs(self.xdtu_0 - other.xdtu_0) <= abserror) and \
            (abs(self.ydtu_0 - other.ydtu_0) <= abserror) and \
            (abs(self.zdtu_0 - other.zdtu_0) <= abserror)
        return result

    def outdict(self, ndigits=3):
        """Return dictionary structure rounded to a given precision."""

        output = self.__dict__.copy()
        for item in output:
            output[item] = round(output[item], ndigits)
        return output


def average_dtu_configurations(list_of_objects):
    """Return DtuConfiguration instance with averaged values.

    Parameters
    ----------
    list_of_objects : python list
        List of DtuConfiguration instances to be averaged.

    Returns
    -------
    result : DtuConfiguration instance
        Object with averaged values.

    """

    result = DtuConfiguration()

    if len(list_of_objects) == 0:
        return result

    list_of_members = result.__dict__.keys()

    # compute average of all the members of the class
    for member in list_of_members:
        result.__dict__[member] = np.mean(
            [tmp_dtu.__dict__[member] for tmp_dtu in list_of_objects]
        )

    return result


def maxdiff_dtu_configurations(list_of_objects):
    """Return DtuConfiguration instance with maximum differences.

    Parameters
    ----------
    list_of_objects : python list
        List of DtuConfiguration instances to be averaged.

    Returns
    -------
    result : DtuConfiguration instance
        Object with averaged values.

    """

    result = DtuConfiguration()

    if len(list_of_objects) == 0:
        return result

    list_of_members = result.__dict__.keys()

    # compute maximum difference for each member
    for member in list_of_members:
        tmp_array = np.array(
            [tmp_dtu.__dict__[member] for tmp_dtu in list_of_objects]
        )
        minval = tmp_array.min()
        maxval = tmp_array.max()
        result.__dict__[member] = maxval - minval

    return result



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
from datetime import datetime
import json
import numpy as np
import os.path
from scipy import ndimage
from scipy.signal import savgol_filter
import socket
import sys
from uuid import uuid4

from numina.array.ccd_line import SpectrumTrail
from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximshow import ximshow

from emirdrp.instrument.csu_configuration import CsuConfiguration
from emirdrp.instrument.dtu_configuration import DtuConfiguration

from emirdrp.core import EMIR_NBARS
from emirdrp.core import EMIR_VALID_GRISMS
from emirdrp.core import EMIR_VALID_FILTERS
from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES


class SlitletLimits(object):
    """Relevant parameters to fit boundaries of a given slitlet.

    Parameters
    ----------
    grism : string
        Grism name. It must be one in EMIR_VALID_GRISMS.
    spfilter : string
        Filter name. It must be one in EMIR_VALID_FILTERS.
    islitlet : int
        Number of slitlet.

    Attributes
    ----------
    grism : string
        Grism name. It must be one in EMIR_VALID_GRISMS.
    spfilter : string
        Filter name. It must be one in EMIR_VALID_FILTERS.
    islitlet : int
        Number of slitlet.
    bb_nc1_orig : int
        Minimum X coordinate (in pixel units) of the rectangle enclosing
        the 2D slitlet image.
    bb_nc2_orig : int
        Maximum X coordinate (in pixel units) of the rectangle enclosing
        the 2D slitlet image.
    bb_ns1_orig : int
        Minimum Y coordinate (in pixel units) of the rectangle enclosing
        the 2D slitlet image.
    bb_ns2_orig : int
        Maximum Y coordinate (in pixel units) of the rectangle enclosing
        the 2D slitlet image.
    xmin_lower_boundary_fit : int
        Minimum abscissa to fit lower boundary.
    xmax_lower_boundary_fit : int
        Maximum abscissa to fit lower boundary.
    xmin_upper_boundary_fit : int
        Minimum abscissa to fit upper boundary.
    xmax_upper_boundary_fit : int
        Maximum abscissa to fit upper boundary.

    """

    def __init__(self, grism, spfilter, islitlet):
        # parameters
        self.grism = grism
        self.spfilter = spfilter
        self.islitlet = islitlet

        # general protections
        if grism not in EMIR_VALID_GRISMS:
            raise ValueError("Grism " + str(grism) + " is not a valid option")
        if spfilter not in EMIR_VALID_FILTERS:
            raise ValueError("Filter " + str(spfilter) +
                             " is not a valid  option")
        if islitlet < 1 or islitlet > EMIR_NBARS:
            raise ValueError("islitlet=" + str(islitlet) +
                             " is outside valid range")

        # check that islitlet is within valid the range for the
        # grism + filter configuration
        if grism == "J" and spfilter == "J":
            if islitlet < 2 or islitlet > 54:
                raise ValueError("islitlet=" + str(islitlet) +
                                 " is outside valid range for grism " +
                                 str(grism) + " and filter " + str(spfilter))
        elif grism == "H" and spfilter == "H":
            if islitlet < 2 or islitlet > 54:
                raise ValueError("islitlet=" + str(islitlet) +
                                 " is outside valid range for grism " +
                                 str(grism) + " and filter " + str(spfilter))
        elif grism == "K" and spfilter == "Ksp":
            if islitlet < 2 or islitlet > 54:
                raise ValueError("islitlet=" + str(islitlet) +
                                 " is outside valid range for grism " +
                                 str(grism) + " and filter " + str(spfilter))
        elif grism == "LR" and spfilter == "YJ":
            if islitlet < 4 or islitlet > 55:
                raise ValueError("islitlet=" + str(islitlet) +
                                 " is outside valid range for grism " +
                                 str(grism) + " and filter " + str(spfilter))
        elif grism == "LR" and spfilter == "HK":
            if islitlet < 4 or islitlet > 55:
                raise ValueError("islitlet=" + str(islitlet) +
                                 " is outside valid range for grism " +
                                 str(grism) + " and filter " + str(spfilter))
        else:
            raise ValueError("Minimum and maximum islitlet still undefined "
                             "for grism " + str(grism) +
                             " and filter " + str(spfilter))

        # expected boundaries of the rectangle enclosing the 2D image
        coeff_bb_ns1 =  [-8.03677111e+01, 3.98169266e+01, -7.77949391e-02,
                         9.00823598e-04]
        delta_bb_ns2 = 84
        # offset measured overplotting ds9_bounddict.reg (created with the
        # script fit_boundaries) corresponding to grism J and filter J
        # over the same slitlet configuration with the corresponding grism
        # and filter setting
        if grism == "J" and spfilter == "J":
            offset_with_J_J = 0
        elif grism == "H" and spfilter == "H":
            offset_with_J_J = 3.0
        elif grism == "K" and spfilter == "Ksp":
            offset_with_J_J = 3.0
        elif grism == "LR" and spfilter == "YJ":
            offset_with_J_J = -95.0
        elif grism == "LR" and spfilter == "HK":
            offset_with_J_J = -95.0
        else:
            raise ValueError("Boundaries still undefined for grism " +
                             str(grism) + " and filter " + str(spfilter))
        coeff_bb_ns1[0] += offset_with_J_J
        poly_bb_ns1 = np.polynomial.Polynomial(coeff_bb_ns1)
        self.bb_nc1_orig = 1
        self.bb_nc2_orig = EMIR_NAXIS1

        self.bb_ns1_orig = int(poly_bb_ns1(islitlet) + 0.5)
        if self.bb_ns1_orig < 1:
            self.bb_ns1_orig = 1
        if self.bb_ns1_orig > EMIR_NAXIS2:
            raise ValueError("Unexpected bb_ns1_orig > EMIR_NAXIS2")
        self.bb_ns2_orig = int(poly_bb_ns1(islitlet) + 0.5) + delta_bb_ns2
        if self.bb_ns2_orig > EMIR_NAXIS2:
            self.bb_ns2_orig = EMIR_NAXIS2
        if self.bb_ns2_orig < 1:
            raise ValueError("Unexpected bb_ns2_orig < 1")

        # X ranges to fit boundaries
        self.xmin_lower_boundary_fit = 1
        self.xmax_lower_boundary_fit = EMIR_NAXIS1
        self.xmin_upper_boundary_fit = 1
        self.xmax_upper_boundary_fit = EMIR_NAXIS1

        # specific corrections for each grism + filter configuration
        if grism == "J" and spfilter == "J":
            self.deg_boundary = 5
            if islitlet == 54:
                self.xmin_upper_boundary_fit = 400
                self.xmax_upper_boundary_fit = 1750
        elif grism == "H" and spfilter == "H":
            self.deg_boundary = 5
            if islitlet == 54:
                self.xmin_upper_boundary_fit = 600
                self.xmax_upper_boundary_fit = 1600
        elif grism == "K" and spfilter == "Ksp":
            self.deg_boundary = 5
            if islitlet == 54:
                self.xmin_upper_boundary_fit = 400
                self.xmax_upper_boundary_fit = 1750
        elif grism == "LR" and spfilter == "YJ":
            self.deg_boundary = 5
            if islitlet == 4:
                self.xmin_lower_boundary_fit = 300
                # self.xmax_lower_boundary_fit = 1750
        elif grism == "LR" and spfilter == "HK":
            self.deg_boundary = 5
            if islitlet == 4:
                self.xmin_lower_boundary_fit = 300
                # self.xmax_lower_boundary_fit = 1750
        else:
            raise ValueError("Ranges to fit boundaries still undefined "
                             "for grism " + str(grism) +
                             " and filter " + str(spfilter))


def extract_slitlet2d(image_2k2k, sltlim):
    """Extract slitlet 2d image from image with original EMIR dimensions.

    Parameters
    ----------
    image_2k2k : 2d numpy array, float
        Original image (dimensions NAXIS1 * NAXIS2)
    sltlim : instance of SlitLimits class
        Object containing relevant information concerning the slitlet
        region to be extracted.

    Returns
    -------
    slitlet2d : 2d numpy array, float
        Image corresponding to the slitlet region defined by its
        bounding box.

    """

    # extract slitlet region
    slitlet2d = image_2k2k[(sltlim.bb_ns1_orig - 1):sltlim.bb_ns2_orig,
                           (sltlim.bb_nc1_orig - 1):sltlim.bb_nc2_orig]

    # transform to float
    slitlet2d = slitlet2d.astype(np.float)

    # return slitlet image
    return slitlet2d


def compute_slitlet_boundaries(
        filename, grism, spfilter, list_slitlets,
        size_x_medfilt, size_y_savgol,
        times_sigma_threshold,
        bounddict, debugplot=0):
    """Compute slitlet boundaries using continuum lamp images.

    Parameters
    ----------
    filename : string
        Input continumm lamp image.
    grism : string
        Grism name. It must be one in EMIR_VALID_GRISMS.
    spfilter : string
        Filter name. It must be one in EMIR_VALID_FILTERS.
    list_slitlets : list of integers
        Number of slitlets to be updated.
    size_x_medfilt : int
        Window in the X (spectral) direction, in pixels, to apply
        the 1d median filter in order to remove bad pixels.
    size_y_savgol : int
        Window in the Y (spatial) direction to be used when using the
        1d Savitzky-Golay filter.
    times_sigma_threshold : float
        Times sigma to detect peaks in derivatives.
    bounddict : dictionary of dictionaries
        Structure to store the boundaries.
    debugplot : int
        Determines whether intermediate computations and/or plots are
        displayed.

    """

    # read 2D image
    hdulist = fits.open(filename)
    image_header = hdulist[0].header
    image2d = hdulist[0].data
    naxis2, naxis1 = image2d.shape
    hdulist.close()
    if debugplot >= 10:
        print('>>> NAXIS1:', naxis1)
        print('>>> NAXIS2:', naxis2)


    # ToDo: replace this by application of cosmetic defect mask!
    for j in range(1024):
        image2d[1024, j] = (image2d[1023, j] + image2d[1025, j]) / 2
        image2d[1023, j + 1024] = (image2d[1022, j + 1024] +
                                   image2d[1024, j + 1024]) / 2
        
    # remove path from filename
    sfilename = os.path.basename(filename)

    # check that the FITS file has been obtained with EMIR
    instrument = image_header['instrume']
    if instrument != 'EMIR':
        raise ValueError("INSTRUME keyword is not 'EMIR'!")

    # read CSU configuration from FITS header
    csu_config = CsuConfiguration.define_from_fits(filename)

    # read DTU configuration from FITS header
    dtu_config = DtuConfiguration.define_from_fits(filename)

    # read grism
    grism_in_header = image_header['grism']
    if grism != grism_in_header:
        raise ValueError("GRISM keyword=" + grism_in_header +
                         " is not the expected value=" + grism)

    # read filter
    spfilter_in_header = image_header['filter']
    if spfilter != spfilter_in_header:
        raise ValueError("FILTER keyword=" + spfilter_in_header +
                         " is not the expected value=" + spfilter)

    # read rotator position angle
    rotang = image_header['rotang']

    # read date-obs
    date_obs = image_header['date-obs']

    for islitlet in list_slitlets:
        if debugplot < 10:
            sys.stdout.write('.')
            sys.stdout.flush()

        sltlim = SlitletLimits(grism, spfilter, islitlet)
        # extract slitlet2d
        slitlet2d = extract_slitlet2d(image2d, sltlim)
        if debugplot % 10 != 0:
            ximshow(slitlet2d,
                    title=sfilename + " [original]"
                          "\nslitlet=" + str(islitlet) +
                          ", grism=" + grism +
                          ", filter=" + spfilter +
                          ", rotang=" + str(round(rotang, 2)),
                    first_pixel=(sltlim.bb_nc1_orig, sltlim.bb_ns1_orig),
                    debugplot=debugplot)

        # apply 1d median filtering (along the spectral direction)
        # to remove bad pixels
        size_x = size_x_medfilt
        size_y = 1
        slitlet2d_smooth = ndimage.filters.median_filter(
            slitlet2d, size=(size_y, size_x))

        if debugplot % 10 != 0:
            ximshow(slitlet2d_smooth,
                    title=sfilename + " [smoothed]"
                          "\nslitlet=" + str(islitlet) +
                          ", grism=" + grism +
                          ", filter=" + spfilter +
                          ", rotang=" + str(round(rotang, 2)),
                    first_pixel=(sltlim.bb_nc1_orig, sltlim.bb_ns1_orig),
                    debugplot=debugplot)

        # apply 1d Savitzky-Golay filter (along the spatial direction)
        # to compute first derivative
        slitlet2d_savgol = savgol_filter(
            slitlet2d_smooth, window_length=size_y_savgol, polyorder=2,
            deriv=1, axis=0)

        # compute basic statistics
        q25, q50, q75 = np.percentile(slitlet2d_savgol, q=[25.0, 50.0, 75.0])
        sigmag = 0.7413 * (q75 - q25)  # robust standard deviation
        if debugplot >= 10:
            print("q50, sigmag:", q50, sigmag)

        if debugplot % 10 != 0:
            ximshow(slitlet2d_savgol,
                    title=sfilename + " [S.-G.filt.]"
                          "\nslitlet=" + str(islitlet) +
                          ", grism=" + grism +
                          ", filter=" + spfilter +
                          ", rotang=" + str(round(rotang, 2)),
                    first_pixel=(sltlim.bb_nc1_orig, sltlim.bb_ns1_orig),
                    z1z2=(q50-times_sigma_threshold*sigmag,
                          q50+times_sigma_threshold*sigmag),
                    debugplot=debugplot)

        # identify objects in slitlet2d_savgol: pixels with positive
        # derivatives are identify independently from pixels with
        # negative derivaties; then the two set of potential features
        # are merged; this approach avoids some problems when, in
        # nearby regions, there are pixels with positive and negative
        # derivatives (in those circumstances a single search as
        # np.logical_or(
        #     slitlet2d_savgol < q50 - times_sigma_threshold * sigmag,
        #     slitlet2d_savgol > q50 + times_sigma_threshold * sigmag)
        # led to erroneous detections!)
        #
        # search for positive derivatives
        labels2d_objects_pos, no_objects_pos = ndimage.label(
            slitlet2d_savgol > q50 + times_sigma_threshold * sigmag)
        # search for negative derivatives
        labels2d_objects_neg, no_objects_neg = ndimage.label(
            slitlet2d_savgol < q50 - times_sigma_threshold * sigmag)
        # merge both sets
        non_zero_neg = np.where(labels2d_objects_neg > 0)
        labels2d_objects = np.copy(labels2d_objects_pos)
        labels2d_objects[non_zero_neg] += \
            labels2d_objects_neg[non_zero_neg] + no_objects_pos
        no_objects = no_objects_pos + no_objects_neg

        if debugplot >= 10:
            print("Number of objects with positive derivative:",
                  no_objects_pos)
            print("Number of objects with negative derivative:",
                  no_objects_neg)
            print("Total number of objects initially found...:", no_objects)

        if debugplot % 10 != 0:
            ximshow(labels2d_objects,
                    z1z2=(0, no_objects),
                    title=sfilename + " [objects]"
                                     "\nslitlet=" + str(islitlet) +
                          ", grism=" + grism +
                          ", filter=" + spfilter +
                          ", rotang=" + str(round(rotang, 2)),
                    first_pixel=(sltlim.bb_nc1_orig, sltlim.bb_ns1_orig),
                    cbar_label="Object number",
                    debugplot=debugplot)

        # select boundaries as the largest objects found with
        # positive and negative derivatives
        n_der_pos = 0  # number of pixels covered by the object with deriv > 0
        i_der_pos = 0  # id of the object with deriv > 0
        n_der_neg = 0  # number of pixels covered by the object with deriv < 0
        i_der_neg = 0  # id of the object with deriv < 0
        for i in range(1, no_objects+1):
            xy_tmp = np.where(labels2d_objects == i)
            n_pix = len(xy_tmp[0])
            if i <= no_objects_pos:
                if n_pix > n_der_pos:
                    i_der_pos = i
                    n_der_pos = n_pix
            else:
                if n_pix > n_der_neg:
                    i_der_neg = i
                    n_der_neg = n_pix

        # determine which boundary is lower and which is upper
        y_center_mass_der_pos = ndimage.center_of_mass(
            slitlet2d_savgol, labels2d_objects, [i_der_pos])[0][0]
        y_center_mass_der_neg = ndimage.center_of_mass(
            slitlet2d_savgol, labels2d_objects, [i_der_neg])[0][0]
        if y_center_mass_der_pos < y_center_mass_der_neg:
            i_lower = i_der_pos
            i_upper = i_der_neg
            if debugplot >= 10:
                print("-> lower boundary has positive derivatives")
        else:
            i_lower = i_der_neg
            i_upper = i_der_pos
            if debugplot >= 10:
                print("-> lower boundary has negative derivatives")
        list_slices_ok = [i_lower, i_upper]

        # adjust individual boundaries passing the selection:
        # - select points in the image belonging to a given boundary
        # - compute weighted mean of the pixels of the boundary, column
        #   by column (this reduces dramatically the number of points
        #   to be fitted to determine the boundary)
        list_boundaries = []
        for k in range(2):  # k=0 lower boundary, k=1 upper boundary
            # select points to be fitted for a particular boundary
            # (note: be careful with array indices and pixel
            # coordinates)
            xy_tmp = np.where(labels2d_objects == list_slices_ok[k])
            xmin = xy_tmp[1].min()  # array indices (integers)
            xmax = xy_tmp[1].max()  # array indices (integers)
            xfit = []
            yfit = []
            # fix range for fit
            if k == 0:
                xmineff = max(sltlim.xmin_lower_boundary_fit, xmin)
                xmaxeff = min(sltlim.xmax_lower_boundary_fit, xmax)
            else:
                xmineff = max(sltlim.xmin_upper_boundary_fit, xmin)
                xmaxeff = min(sltlim.xmax_upper_boundary_fit, xmax)
            # loop in columns of the image belonging to the boundary
            for xdum in range(xmineff, xmaxeff + 1):  # array indices (integer)
                iok = np.where(xy_tmp[1] == xdum)
                y_tmp = xy_tmp[0][iok] + sltlim.bb_ns1_orig  # image pixel
                weight = slitlet2d_savgol[xy_tmp[0][iok], xy_tmp[1][iok]]
                y_wmean = sum(y_tmp * weight) / sum(weight)
                xfit.append(xdum + sltlim.bb_nc1_orig)
                yfit.append(y_wmean)
            xfit = np.array(xfit)
            yfit = np.array(yfit)
            # declare new SpectrumTrail instance
            boundary = SpectrumTrail()
            # define new boundary
            boundary.fit(x=xfit, y=yfit, deg=sltlim.deg_boundary,
                         times_sigma_reject=10,
                         title="slit:" + str(sltlim.islitlet) +
                               ", deg=" + str(sltlim.deg_boundary),
                         debugplot=0)
            list_boundaries.append(boundary)

        if debugplot % 10 != 0:
            for tmp_img, tmp_label in zip(
                [slitlet2d_savgol, slitlet2d],
                [' [S.-G.filt.]', ' [original]']
            ):
                ax = ximshow(tmp_img,
                             title=sfilename + tmp_label +
                                   "\nslitlet=" + str(islitlet) +
                                   ", grism=" + grism +
                                   ", filter=" + spfilter +
                                   ", rotang=" + str(round(rotang, 2)),
                             first_pixel=(sltlim.bb_nc1_orig,
                                          sltlim.bb_ns1_orig),
                             show=False,
                             debugplot=debugplot)
                for k in range(2):
                    xpol, ypol = list_boundaries[k].linspace_pix(
                        start=1, stop=EMIR_NAXIS1)
                    ax.plot(xpol, ypol, 'b--', linewidth=1)
                for k in range(2):
                    xpol, ypol = list_boundaries[k].linspace_pix()
                    ax.plot(xpol, ypol, 'g--', linewidth=4)
                # show plot
                pause_debugplot(debugplot, pltshow=True)

        # update bounddict
        tmp_dict = {
            'boundary_coef_lower':
                list_boundaries[0].poly_funct.coef.tolist(),
            'boundary_xmin_lower': list_boundaries[0].xlower_line,
            'boundary_xmax_lower': list_boundaries[0].xupper_line,
            'boundary_coef_upper':
                list_boundaries[1].poly_funct.coef.tolist(),
            'boundary_xmin_upper': list_boundaries[1].xlower_line,
            'boundary_xmax_upper': list_boundaries[1].xupper_line,
            'csu_bar_left': csu_config.csu_bar_left(islitlet),
            'csu_bar_right': csu_config.csu_bar_right(islitlet),
            'csu_bar_slit_center': csu_config.csu_bar_slit_center(islitlet),
            'csu_bar_slit_width': csu_config.csu_bar_slit_width(islitlet),
            'rotang': rotang,
            'xdtu': dtu_config.xdtu,
            'ydtu': dtu_config.ydtu,
            'zdtu': dtu_config.zdtu,
            'xdtu_0': dtu_config.xdtu_0,
            'ydtu_0': dtu_config.ydtu_0,
            'zdtu_0': dtu_config.zdtu_0,
            'zzz_info1': os.getlogin() + '@' + socket.gethostname(),
            'zzz_info2': datetime.now().isoformat()
        }
        slitlet_label = "slitlet" + str(islitlet).zfill(2)
        if slitlet_label not in bounddict['contents']:
            bounddict['contents'][slitlet_label] = {}
        bounddict['contents'][slitlet_label][date_obs] = tmp_dict

    if debugplot < 10:
        print("")


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()

    # positional arguments
    parser.add_argument("filename",
                        help="FITS file or txt file with list of FITS files",
                        type=argparse.FileType('rb'))
    parser.add_argument("--grism", required=True,
                        help="Grism name",
                        choices=EMIR_VALID_GRISMS)
    parser.add_argument("--filter", required=True,
                        help="Filter name",
                        choices=EMIR_VALID_FILTERS)
    parser.add_argument("--tuple_slit_numbers", required=True,
                        help="Tuple n1[,n2[,step]] to define slitlet numbers")

    # optional arguments
    parser.add_argument("--first_time",
                        help="Generate new bounddict json file",
                        action="store_true")
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting/debugging" +
                             " (default=0)",
                        type=int, default=0,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args()

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # read slitlet numbers to be computed
    tmp_str = args.tuple_slit_numbers.split(",")
    if len(tmp_str) == 3:
        n1 = int(tmp_str[0])
        n2 = int(tmp_str[1])
        step = int(tmp_str[2])
    elif len(tmp_str) == 2:
        n1 = int(tmp_str[0])
        n2 = int(tmp_str[1])
        step = 1
    elif len(tmp_str) == 1:
        n1 = int(tmp_str[0])
        n2 = n1
        step = 1
    else:
        raise ValueError("Invalid tuple for slitlet numbers")
    if n1 < 1:
        raise ValueError("Invalid slitlet number < 1")
    if n2 > EMIR_NBARS:
        raise ValueError("Invalid slitlet number > EMIR_NBARS")
    if step <= 0:
        raise ValueError("Invalid step <= 0")
    list_slitlets = range(n1, n2+1, step)

    # define bounddict file name
    bounddict_file = "bounddict_grism_" + args.grism + "_filter_" + \
                     args.filter + ".json"

    # define bounddict prior to the new computation
    if args.first_time:
        bounddict = {}
        bounddict['instrument'] = 'EMIR'
        bounddict['meta_info'] = {}
        bounddict['meta_info']['creation_date'] = datetime.now().isoformat()
        bounddict['meta_info']['description'] = \
            'slitlet boundaries from fits to continuum-lamp exposures'
        bounddict['meta_info']['recipe_name'] = 'undefined'
        bounddict['tags'] = {}
        bounddict['tags']['grism'] = args.grism
        bounddict['tags']['filter'] = args.filter
        bounddict['uuid'] = str(uuid4())
        bounddict['contents'] = {}
        print('>>> Generating new bounddict from scratch')
    else:
        if not os.path.isfile(bounddict_file):
            raise ValueError("File " + bounddict_file + " not found!")
        else:
            bounddict = json.loads(open(bounddict_file, mode='rt').read())
            print('>>> Initializing bounddict from previous file:')
            print(bounddict_file)

    # if input file is a txt file, assume it is a list of FITS files
    filename = args.filename.name
    if filename[-4:] == ".txt":
        # since the input filename has been opened with argparse in binary
        # mode, it is necessary to close it and open it in text mode
        args.filename.close()
        with open(filename, mode='rt') as f:
            file_content = f.read().splitlines()
        list_fits_files = []
        for line in file_content:
            if len(line) > 0:
                if line[0] != '#':
                    tmpfile = line.split()[0]
                    if not os.path.isfile(tmpfile):
                        raise ValueError("File " + tmpfile + " not found!")
                    list_fits_files.append(tmpfile)
    else:
        list_fits_files = [filename]

    # update bounddict
    for ifile, myfile in enumerate(list_fits_files):
        print('>>> Reading file ' + str(ifile+1) + "/" +
              str(len(list_fits_files)) + ":\n" + myfile)
        compute_slitlet_boundaries(
            filename=myfile,
            list_slitlets=list_slitlets,
            grism=args.grism, spfilter=args.filter,
            size_x_medfilt=31,
            size_y_savgol=11,
            times_sigma_threshold=3.0,
            bounddict=bounddict, debugplot=args.debugplot)

    # save new version of bounddict
    # print(json.dumps(bounddict, indent=4, sort_keys=True))
    with open(bounddict_file, 'wt') as fstream:
        json.dump(bounddict, fstream, indent=2, sort_keys=True)
        print('>>> Saving file ' + bounddict_file)


if __name__ == "__main__":

    main()



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


"""
Spectroscopy mode, Stare Spectra
"""


from numina.core import Result
from numina.array.combine import median
from numina.processing.combine import basic_processing_with_combination

from emirdrp.core.recipe import EmirRecipe
import emirdrp.requirements as reqs
import emirdrp.products as prods
from emirdrp.processing.wavecal.apply_rectwv_coeff import apply_rectwv_coeff
from emirdrp.processing.wavecal.rectwv_coeff_from_mos_library \
    import rectwv_coeff_from_mos_library
from emirdrp.processing.wavecal.rectwv_coeff_to_ds9 import save_four_ds9


class SkySpecRecipe(EmirRecipe):
    """Recipe to process data taken in spectral sky mode.

    """

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterSpectralFlatFieldRequirement()
    master_rectwv = reqs.MasterRectWaveRequirement()
    skyspec = Result(prods.SkySpectrum)
    reduced_image = Result(prods.ProcessedMOS)

    def run(self, rinput):
        self.logger.info('starting spectral sky reduction')

        flow = self.init_filters(rinput)

        reduced_image = basic_processing_with_combination(
            rinput, flow,
            method=median,
            errors=True
        )

        hdr = reduced_image[0].header
        self.set_base_headers(hdr)

        # save intermediate image in work directory
        self.save_intermediate_img(reduced_image, 'reduced_image.fits')

        # RectWaveCoeff object with rectification and wavelength calibration
        # coefficients for the particular CSU configuration
        rectwv_coeff = rectwv_coeff_from_mos_library(
            reduced_image,
            rinput.master_rectwv
        )
        # save as JSON file in work directory
        self.save_structured_as_json(rectwv_coeff, 'rectwv_coeff.json')

        # generate associated ds9 region files and save them in work directory
        if self.intermediate_results:
            save_four_ds9(rectwv_coeff)

        # apply rectification and wavelength calibration
        skyspec = apply_rectwv_coeff(
            reduced_image,
            rectwv_coeff
        )

        self.logger.info('end sky spectral reduction')

        result = self.create_result(
            reduced_image=reduced_image,
            skyspec=skyspec
        )

        return result



#
# Copyright 2013-2014 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''Tests for AIV Recipes for EMIR'''

import os
from tempfile import mkstemp
import unittest

from astropy.io import fits
import numpy as np

from numina.core.oresult import ObservationResult, dataframe_from_list
from numina.types.dataframe import DataFrame

from . import SimpleBiasRecipe


class MMTestCase(unittest.TestCase):

    def test1(self):

        # Create some HDUList in memory
        somefits = []
        nimg = 10
        for i in range(nimg):
            hdu = fits.PrimaryHDU(data=np.zeros((10, 10), dtype='int16'))
            hdul = fits.HDUList([hdu])
            somefits.append(hdul)

        #
        # Create the recipe_input
        obsres = ObservationResult()
        obsres.frames = [dataframe_from_list(val) for val in somefits]
        # empty requirements
        reqs = SimpleBiasRecipe.RecipeInput(obresult=obsres)
        recipe = SimpleBiasRecipe()
        #
        result = recipe(reqs)

        self.verify_result(result)

    def verify_result(self, result):

        self.assertIsInstance(result, SimpleBiasRecipe.RecipeResult)
        self.assertTrue(hasattr(result, "biasframe"))
        self.assertIsInstance(result.biasframe, DataFrame)

        frame = result.biasframe.frame

        self.assertIsInstance(frame, fits.HDUList)

        # Number of extensions
        self.assertEqual(len(frame), 3)

        self.assertIsInstance(frame[0], fits.PrimaryHDU)
        self.assertIsInstance(frame[1], fits.ImageHDU)
        self.assertIsInstance(frame[2], fits.ImageHDU)

    def test3(self):

        # Create some HDUList
        somefits = []
        nimg = 10
        for i in range(nimg):
            hdu = fits.PrimaryHDU(data=np.zeros((10, 10), dtype='int16'))
            hdul = fits.HDUList([hdu])
            fd, filename = mkstemp()
            hdul.writeto(filename, overwrite=True)
            somefits.append(filename)

        #
        # Create the recipe_input
        obsres = ObservationResult()
        obsres.frames = [dataframe_from_list(val) for val in somefits]
        # empty requirements
        reqs = SimpleBiasRecipe.RecipeInput(obresult=obsres)
        #
        recipe = SimpleBiasRecipe()
        result = recipe(reqs)

        for fname in somefits:
            os.remove(fname)

        self.verify_result(result)



#
# Copyright 2008-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
import numpy as np
import sys

import emirdrp.datamodel as datamodel
from emirdrp.instrument.csu_configuration import CsuConfiguration

from numina.array.display.ximshow import ximshow
from numina.array.wavecalib.fix_pix_borders import define_mask_borders
from numina.frame.utils import copy_img
from numina.tools.arg_file_is_new import arg_file_is_new

from emirdrp.core import EMIR_NBARS
from emirdrp.core import EMIR_NPIXPERSLIT_RECTIFIED

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES
from emirdrp.core import EMIR_MINIMUM_SLITLET_WIDTH_MM
from emirdrp.core import EMIR_MAXIMUM_SLITLET_WIDTH_MM


def median_slitlets_rectified(
        input_image,
        mode=0,
        list_useful_slitlets=None,
        debugplot=0
):
    """Compute median spectrum for each slitlet.

    Parameters
    ----------
    input_image : HDUList object
        Input 2D image.
    mode : int
        Indicate desired result:
        0 : image with the same size as the input image, with the
            median spectrum of each slitlet spanning all the spectra
            of the corresponding slitlet
        1 : image with 55 spectra, containing the median spectra of
            each slitlet
        2 : single collapsed median spectrum, using exclusively the
            useful slitlets from the input image
    list_useful_slitlets : list of integers or None
        List of useful slitlets (from 1 to EMIR_NBARS). If None, the
        list contains all the slitlets.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    image_median : HDUList object
        Output image.

    """

    result = copy_img(input_image)
    image_header = input_image[0].header
    image2d = input_image[0].data

    # check image dimensions
    naxis2_expected = EMIR_NBARS * EMIR_NPIXPERSLIT_RECTIFIED

    naxis2, naxis1 = image2d.shape
    if naxis2 != naxis2_expected:
        raise ValueError("NAXIS2={0} should be {1}".format(
            naxis2, naxis2_expected
        ))

    # check that the FITS file has been obtained with EMIR
    instrument = image_header['instrume']
    if instrument != 'EMIR':
        raise ValueError("INSTRUME keyword is not 'EMIR'!")

    # initialize output image
    if mode == 0:
        image2d_median = np.zeros((naxis2, naxis1), dtype='float32')
    else:
        image2d_median = np.zeros((EMIR_NBARS, naxis1), dtype='float32')

    # main loop
    for i in range(EMIR_NBARS):
        ns1 = i * EMIR_NPIXPERSLIT_RECTIFIED + 1
        ns2 = ns1 + EMIR_NPIXPERSLIT_RECTIFIED - 1
        sp_median = np.median(image2d[(ns1-1):ns2, :], axis=0)

        if mode == 0:
            image2d_median[(ns1-1):ns2, :] = np.tile(
                sp_median, (EMIR_NPIXPERSLIT_RECTIFIED, 1)
            )
        else:
            image2d_median[i] = np.copy(sp_median)

    if mode == 2:
        # define wavelength calibration parameters
        crpix1 = image_header['crpix1']
        crval1 = image_header['crval1']
        cdelt1 = image_header['cdelt1']

        # segregate slitlets
        if list_useful_slitlets is None:
            list_useful_slitlets = list(range(1, EMIR_NBARS + 1))
            list_not_useful_slitlets = []
        else:
            list_not_useful_slitlets = [
                i for i in list(range(1, EMIR_NBARS + 1))
                if i not in list_useful_slitlets
            ]
        if abs(debugplot) != 0:
            print('>>> list_useful_slitlets....:', list_useful_slitlets)
            print('>>> list_not_useful_slitlets:', list_not_useful_slitlets)

        # define mask from array data
        mask2d, borders = define_mask_borders(image2d_median, sought_value=0)
        if abs(debugplot) % 10 != 0:
            ximshow(mask2d.astype(int), z1z2=(-.2, 1.2), crpix1=crpix1,
                    crval1=crval1, cdelt1=cdelt1, debugplot=debugplot)

        # update mask with unused slitlets
        for islitlet in list_not_useful_slitlets:
            mask2d[islitlet - 1, :] = np.array([True] * naxis1)
        if abs(debugplot) % 10 != 0:
            ximshow(mask2d.astype(int), z1z2=(-.2, 1.2), crpix1=crpix1,
                    crval1=crval1, cdelt1=cdelt1, debugplot=debugplot)

        # useful image pixels
        image2d_masked = image2d_median * (1 - mask2d.astype(int))
        if abs(debugplot) % 10 != 0:
            ximshow(image2d_masked, crpix1=crpix1, crval1=crval1,
                    cdelt1=cdelt1, debugplot=debugplot)

        # masked image
        image2d_masked = np.ma.masked_array(image2d_median, mask=mask2d)
        # median spectrum
        image1d_median = np.ma.median(image2d_masked, axis=0).data

        result[0].data = image1d_median.astype('float32')
    else:
        result[0].data = image2d_median.astype('float32')

    return result


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: compute median spectrum for each slitlet'
    )

    # positional arguments
    parser.add_argument("fitsfile",
                        help="Input FITS file name",
                        type=argparse.FileType('rb'))
    parser.add_argument("outfile",
                        help="Output FITS file name",
                        type=lambda x: arg_file_is_new(parser, x, mode='wb'))

    # optional arguments
    parser.add_argument("--mode",
                        help="Output type: 0 -> full frame (default), "
                             "1 -> individual slitlets, "
                             "2 -> collapsed single spectrum)",
                        default=0, type=int,
                        choices=[0, 1, 2])
    parser.add_argument("--minimum_slitlet_width_mm",
                        help="Minimum slitlet width (mm) for --mode 2 "
                             "(default=0)",
                        default=EMIR_MINIMUM_SLITLET_WIDTH_MM, type=float)
    parser.add_argument("--maximum_slitlet_width_mm",
                        help="Maximum slitlet width (mm) for --mode 2 "
                             "(default=" +
                             str(EMIR_MAXIMUM_SLITLET_WIDTH_MM) + ")",
                        default=EMIR_MAXIMUM_SLITLET_WIDTH_MM, type=float)
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting/debugging" +
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # read input FITS file
    hdulist = fits.open(args.fitsfile)

    # determine useful slitlets
    mecs_header = datamodel.get_mecs_header(hdulist)
    csu_config = CsuConfiguration.define_from_header(mecs_header)
    # segregate slitlets
    list_useful_slitlets = csu_config.widths_in_range_mm(
        minwidth=args.minimum_slitlet_width_mm,
        maxwidth=args.maximum_slitlet_width_mm
    )

    image_median = median_slitlets_rectified(
        hdulist,
        mode=args.mode,
        list_useful_slitlets=list_useful_slitlets,
        debugplot=args.debugplot
    )

    # save result
    image_median.writeto(args.outfile, overwrite=True)


if __name__ == "__main__":

    main()



import pytest
import astropy.io.fits as fits
import numpy

import numina.core
import numina.exceptions

from ..stare import StareImageBaseRecipe


def create_frame(val=0, keys=None):
    data = numpy.zeros((10, 10)) + val
    hdu = fits.PrimaryHDU(data)
    if keys:
        for k, v in keys.items():
            hdu.header[k] = v

    return fits.HDUList([hdu])


def create_frame_bpm():
    data = numpy.zeros((10, 10), dtype='uint8')
    hdu = fits.PrimaryHDU(data)
    hdu.header['UUID'] = '74263122-4747-4a68-b148-7834d702a733'
    return fits.HDUList([hdu])


def create_frame_dark():
    data = numpy.zeros((10, 10))
    hdu = fits.PrimaryHDU(data)
    hdu.header['UUID'] = 'db4cf732-76a7-4f54-85bf-2db2f0ab46ef'
    return fits.HDUList([hdu])


def create_frame_flat():
    data = numpy.ones((10, 10))
    hdu = fits.PrimaryHDU(data)
    hdu.header['UUID'] = 'f7af05ee-27bd-4a04-957d-28d23b11a532'
    return fits.HDUList([hdu])


def create_ob(value, nimages, exptime=100.0, starttime=0.0):

    frames = []
    for i in range(nimages):
        base = starttime
        off1 = i * exptime
        off2 = off1 + exptime
        t1 = base + off1
        t2 = base + off2
        keys = {'DATE-OBS': '2019-04-12T03:01:05.678', 'TSUTC1': t1, 'TSUTC2': t2}
        frame = numina.core.DataFrame(frame=create_frame(val=value, keys=keys))
        frames.append(frame)

    obsresult = numina.core.ObservationResult()
    obsresult.frames = frames
    return obsresult


@pytest.mark.parametrize("nimages", [1,2,3,4])
def test_stare_with_bpm(nimages):

    value = 4
    starttime = 1003000001.00034
    exptime = 102.0
    expected_data = value * numpy.ones((10, 10))
    obsresult = create_ob(value, nimages, exptime, starttime)

    recipe = StareImageBaseRecipe()
    rinput = recipe.create_input(
        obresult=obsresult,
        master_bpm=numina.core.DataFrame(frame= create_frame_bpm()),
        master_dark=numina.core.DataFrame(frame=create_frame_dark()),
        master_flat=numina.core.DataFrame(frame=create_frame_flat())
    )

    result = recipe.run(rinput)
    hdulist = result.frame.open()

    assert 'BPM' in hdulist

    hdu = hdulist[0]
    assert hdu.header['NUM-NCOM'] == nimages
    assert hdu.header['NUM-BPM'] == 'uuid:74263122-4747-4a68-b148-7834d702a733'
    assert hdu.header['NUM-DK'] == 'uuid:db4cf732-76a7-4f54-85bf-2db2f0ab46ef'
    assert hdu.header['NUM-FF'] == 'uuid:f7af05ee-27bd-4a04-957d-28d23b11a532'
    assert hdu.header['TSUTC1'] == starttime
    assert hdu.header['TSUTC2'] == starttime + nimages * exptime
    assert numpy.allclose(expected_data, hdu.data)


@pytest.mark.parametrize("nimages", [1,2,3,4])
def test_stare(nimages):

    value = 9
    starttime = 1030040001.00034
    exptime = 105.0
    expected_data = value * numpy.ones((10, 10))
    obsresult = create_ob(value, nimages, exptime, starttime)

    recipe = StareImageBaseRecipe()
    rinput = recipe.create_input(
        obresult=obsresult,
        master_dark=numina.core.DataFrame(frame=create_frame_dark()),
        master_flat=numina.core.DataFrame(frame=create_frame_flat())
    )

    result = recipe.run(rinput)
    hdulist = result.frame.open()

    bpm_ext = 0
    try:
        bpm_ext = hdulist['BPM']
    except KeyError:
        pass

    assert bpm_ext != 0, "Wrong BPM extension number"

    hdu = hdulist[0]
    assert hdu.header['NUM-NCOM'] == nimages
    assert hdu.header['NUM-DK'] == 'uuid:db4cf732-76a7-4f54-85bf-2db2f0ab46ef'
    assert hdu.header['NUM-FF'] == 'uuid:f7af05ee-27bd-4a04-957d-28d23b11a532'
    assert hdu.header['TSUTC1'] == starttime
    assert hdu.header['TSUTC2'] == starttime + nimages * exptime
    assert numpy.allclose(expected_data, hdu.data)



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

from emirdrp.core import EMIR_NAXIS2


def nscan_minmax_frontiers(y0_frontier_lower, y0_frontier_upper,
                           resize=False):
    """Compute valid scan range for provided y0_frontier values.
    Parameters
    ----------
    y0_frontier_lower : float
        Ordinate of the lower frontier.
    y0_frontier_upper : float
        Ordinate of the upper frontier.
    resize : bool
        If True, when the limits are beyond the expected values
        [1,EMIR_NAXIS2], the values are truncated.
    Returns
    -------
    nscan_min : int
        Minimum useful scan for the image.
    nscan_max : int
        Maximum useful scan for the image.
    """

    fraction_pixel = y0_frontier_lower - int(y0_frontier_lower)
    if fraction_pixel > 0.0:
        nscan_min = int(y0_frontier_lower) + 1
    else:
        nscan_min = int(y0_frontier_lower)
    if nscan_min < 1:
        if resize:
            nscan_min = 1
        else:
            raise ValueError("nscan_min=" + str(nscan_min) + " is < 1")

    fraction_pixel = y0_frontier_upper - int(y0_frontier_upper)
    if fraction_pixel > 0.0:
        nscan_max = int(y0_frontier_upper)
    else:
        nscan_max = int(y0_frontier_upper) - 1
    if nscan_max > EMIR_NAXIS2:
        if resize:
            nscan_max = EMIR_NAXIS2
        else:
            raise ValueError("nscan_max=" + str(nscan_max) +
                             " is > NAXIS2_EMIR=" + str(EMIR_NAXIS2))

    return nscan_min, nscan_max



#
# Copyright 2015-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Bar characterization using gradients for EMIR"""

from __future__ import division

import numpy
from numina.array.utils import coor_to_pix_1d
from numina.core import Requirement, Result, Parameter
import numina.exceptions
import numina.types.array as tarray
from numpy.polynomial.polynomial import polyval
from scipy.ndimage import convolve1d
from scipy.ndimage.filters import median_filter

from numina.processing.combine import basic_processing_with_combination

import emirdrp.datamodel as datamodel
import emirdrp.products as prods
import emirdrp.requirements as reqs
from emirdrp.core import EMIR_PIXSCALE, EMIR_NBARS, EMIR_RON
from emirdrp.core.recipe import EmirRecipe
from emirdrp.processing.bardetect import char_bar_peak_l, char_bar_peak_r, char_bar_height


class MaskImagingRecipe(EmirRecipe):

    # Recipe Requirements
    #
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = reqs.MasterSkyRequirement()

    bars_nominal_positions = Requirement(prods.CoordinateList2DType,
                                         'Nominal positions of the bars'
                                         )
    median_filter_size = Parameter(5, 'Size of the median box')
    average_box_row_size = Parameter(7, 'Number of rows to average for fine centering (odd)')
    average_box_col_size = Parameter(21, 'Number of columns to extract for fine centering (odd)')
    fit_peak_npoints = Parameter(3, 'Number of points to use for fitting the peak (odd)')

    # Recipe Products
    frame = Result(prods.ProcessedImage)
    # derivative = Result(prods.ProcessedImage)
    slits = Result(tarray.ArrayType)
    positions3 = Result(tarray.ArrayType)
    positions5 = Result(tarray.ArrayType)
    positions7 = Result(tarray.ArrayType)
    positions9 = Result(tarray.ArrayType)
    DTU = Result(tarray.ArrayType)
    ROTANG = Result(float)
    TSUTC1 = Result(float)
    csupos = Result(tarray.ArrayType)
    csusens = Result(tarray.ArrayType)

    def run(self, rinput):
        self.logger.info('starting processing for bars detection')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow=flow)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        try:
            rotang = hdr['ROTANG']
            tsutc1 = hdr['TSUTC1']
            mecs_hdr = datamodel.get_mecs_header(hdulist)
            dtub, dtur = datamodel.get_dtur_from_header(mecs_hdr)
            csupos = datamodel.get_csup_from_header(mecs_hdr)
            csusens = datamodel.get_cs_from_header(mecs_hdr)

        except KeyError as error:
            self.logger.error(error)
            raise numina.exceptions.RecipeError(error)

        self.logger.debug('finding bars')
        # Processed array
        arr = hdulist[0].data

        # Median filter of processed array (two times)
        mfilter_size = rinput.median_filter_size

        self.logger.debug('median filtering X, %d columns', mfilter_size)
        arr_median = median_filter(arr, size=(1, mfilter_size))
        self.logger.debug('median filtering X, %d rows', mfilter_size)
        arr_median = median_filter(arr_median, size=(mfilter_size, 1))

        # Median filter of processed array (two times) in the other direction
        # for Y coordinates
        self.logger.debug('median filtering Y, %d rows', mfilter_size)
        arr_median_alt = median_filter(arr, size=(mfilter_size, 1))
        self.logger.debug('median filtering Y, %d columns', mfilter_size)
        arr_median_alt = median_filter(arr_median_alt, size=(1, mfilter_size))

        xfac = dtur[0] / EMIR_PIXSCALE
        yfac = -dtur[1] / EMIR_PIXSCALE

        vec = [yfac, xfac]
        self.logger.debug('DTU shift is %s', vec)

        # and the table of approx positions of the slits
        barstab = rinput.bars_nominal_positions
        # Currently, we only use fields 0 and 2
        # of the nominal positions file

        # Number or rows used
        # These other parameters cab be tuned also
        bstart = 1
        bend = 2047
        self.logger.debug('ignoring columns outside %d - %d',bstart, bend-1)

        # extract a region to average
        wy = (rinput.average_box_row_size // 2)
        wx = (rinput.average_box_col_size // 2)
        self.logger.debug('extraction window is %d rows, %d cols',2*wy+1, 2*wx+1)
        # Fit the peak with these points
        wfit = 2 * (rinput.fit_peak_npoints // 2) + 1
        self.logger.debug('fit with %d points', wfit)

        # Minimum threshold
        threshold = 5 * EMIR_RON
        # Savitsky and Golay (1964) filter to compute the X derivative
        # scipy >= xx has a savgol_filter function
        # for compatibility we do it manually

        allpos = {}
        ypos3_kernel = None
        slits = numpy.zeros((EMIR_NBARS, 8), dtype='float')

        self.logger.info('start finding bars')
        for ks in [3, 5, 7, 9]:
            self.logger.debug('kernel size is %d', ks)
            # S and G kernel for derivative
            kw = ks * (ks*ks-1) / 12.0
            coeffs_are = -numpy.arange((1-ks)//2, (ks-1)//2 + 1) / kw
            if ks == 3:
                ypos3_kernel = coeffs_are
            self.logger.debug('kernel weights are %s', coeffs_are)

            self.logger.debug('derive image in X direction')
            arr_deriv = convolve1d(arr_median, coeffs_are, axis=-1)
            # Axis 0 is
            #
            self.logger.debug('derive image in Y direction (with kernel=3)')
            arr_deriv_alt = convolve1d(arr_median_alt, ypos3_kernel, axis=0)

            positions = []
            for coords in barstab:
                lbarid = int(coords[0])
                rbarid = lbarid + EMIR_NBARS
                ref_y_coor = coords[1] + vec[1]
                poly_coeffs = coords[2:]
                prow = coor_to_pix_1d(ref_y_coor) - 1
                fits_row = prow + 1 # FITS pixel index

                # A function that returns the center of the bar
                # given its X position
                def center_of_bar(x):
                    # Pixel values are 0-based
                    return polyval(x+1-vec[0], poly_coeffs) + vec[1] - 1

                self.logger.debug('looking for bars with ids %d - %d', lbarid, rbarid)
                self.logger.debug('reference y position is Y %7.2f', ref_y_coor)

                # if ref_y_coor is outlimits, skip this bar
                # ref_y_coor is in FITS format
                if (ref_y_coor >= 2047) or (ref_y_coor <= 1):
                    self.logger.debug('reference y position is outlimits, skipping')
                    positions.append([lbarid, fits_row, fits_row, 1, 0, 3])
                    positions.append([rbarid, fits_row, fits_row, 1, 0, 3])
                    continue

                # Left bar
                self.logger.debug('measure left border (%d)', lbarid)

                centery, xpos, fwhm, st = char_bar_peak_l(arr_deriv, prow, bstart, bend, threshold,
                                                          center_of_bar, wx=wx, wy=wy, wfit=wfit)
                xpos1 = xpos
                positions.append([lbarid, centery+1, fits_row, xpos+1, fwhm, st])

                # Right bar
                self.logger.debug('measure rigth border (%d)', rbarid)
                centery, xpos, fwhm, st = char_bar_peak_r(arr_deriv, prow, bstart, bend, threshold,
                                                          center_of_bar, wx=wx, wy=wy, wfit=wfit)
                positions.append([rbarid, centery+1, fits_row, xpos+1, fwhm, st])
                xpos2 = xpos
                #
                if st == 0:
                    self.logger.debug('measure top-bottom borders')
                    try:
                        y1, y2, statusy = char_bar_height(arr_deriv_alt, xpos1, xpos2, centery, threshold,
                                                          wh=35, wfit=wfit)
                    except Exception as error:
                        self.logger.warning('Error computing height: %s', error)
                        statusy = 44

                    if statusy in [0, 40]:
                        # Main border is detected
                        positions[-1][1] = y2 + 1
                        positions[-2][1] = y2 + 1
                    else:
                        # Update status
                        positions[-1][-1] = 4
                        positions[-2][-1] = 4
                else:
                    self.logger.debug('slit is not complete')
                    y1, y2 = 0, 0

                # Update positions

                self.logger.debug('bar %d centroid-y %9.4f, row %d x-pos %9.4f, FWHM %6.3f, status %d', *positions[-2])
                self.logger.debug('bar %d centroid-y %9.4f, row %d x-pos %9.4f, FWHM %6.3f, status %d', *positions[-1])

                if ks == 5:
                    slits[lbarid - 1] = [xpos1, y2, xpos2, y2, xpos2, y1, xpos1, y1]
                    # FITS coordinates
                    slits[lbarid - 1] += 1.0
                    self.logger.debug('inserting bars %d-%d into "slits"', lbarid, rbarid)

            allpos[ks] = numpy.asarray(positions, dtype='float') # GCS doesn't like lists of lists

        self.logger.debug('end finding bars')
        result = self.create_result(frame=hdulist,
                                    slits=slits,
                                    positions9=allpos[9],
                                    positions7=allpos[7],
                                    positions5=allpos[5],
                                    positions3=allpos[3],
                                    DTU=dtub,
                                    ROTANG=rotang,
                                    TSUTC1=tsutc1,
                                    csupos=csupos,
                                    csusens=csusens,
                                    )
        return result



#
# Copyright 2011-2014 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''

Naming intermediate images

'''


def name_redimensioned_frames(label, step, ext='.fits'):
    dn = '%s_r%s' % (label, ext)
    mn = '%s_mr%s' % (label, ext)
    return dn, mn


def name_object_mask(label, step, ext='.fits'):
    return '%s_mro_i%01d%s' % (label, step, ext)


def name_skybackground(label, step, ext='.fits'):
    dn = '%s_sky_i%01d%s' % (label, step, ext)
    return dn


def name_skybackgroundmask(label, step, ext='.fits'):
    dn = '%s_skymask_i%01d%s' % (label, step, ext)
    return dn


def name_skysub_proc(label, step, ext='.fits'):
    dn = '%s_rfs_i%01d%s' % (label, step, ext)
    return dn


def name_skyflat(label, step, ext='.fits'):
    dn = 'superflat_%s_i%01d%s' % (label, step, ext)
    return dn


def name_skyflat_proc(label, step, ext='.fits'):
    dn = '%s_rf_i%01d%s' % (label, step, ext)
    return dn


def name_segmask(step, ext='.fits'):
    return "check_i%01d%s" % (step, ext)



#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Definition of Slitlet2D class"""

from __future__ import division, print_function

import numpy as np

from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximshow import ximshow
from numina.array.distortion import order_fmap
from numina.array.distortion import rectify2d

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2
from emirdrp.core import EMIR_NPIXPERSLIT_RECTIFIED


class Slitlet2D(object):
    """Slitlet2D class definition.

    Parameters
    ----------
    islitlet : int
        Slitlet number.
    rectwv_coeff : instance of RectWaveCoeff
        Object storing the JSON input file where the
        the rectification and wavelength calibration transformations
        for a particular instrument configuration are stored.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Attributes
    ----------
    islitlet : int
        Slitlet number.
    csu_bar_left : float
        Location (mm) of the left bar for each slitlet.
    csu_bar_right : list of floats
        Location (mm) of the right bar for each slitlet, using the
        same origin employed for csu_bar_left (which is not the
        value stored in the FITS keywords.
    csu_bar_slit_center : list of floats
        Middle point (mm) in between the two bars defining a slitlet.
    csu_bar_slit_width : list of floats
        Slitlet width (mm), computed as the distance between the two
        bars defining the slitlet.
    bb_nc1_orig : int
        Minimum X coordinate of the enclosing bounding box (in pixel
        units) in the original image.
    bb_nc2_orig : int
        Maximum X coordinate of the enclosing bounding box (in pixel
        units) in the original image.
    bb_ns1_orig : int
        Minimum Y coordinate of the enclosing bounding box (in pixel
        units) in the original image.
    bb_ns2_orig : int
        Maximum Y coordinate of the enclosing bounding box (in pixel
        units) in the original image.
    x0_reference : float
        X coordinate where the rectified y0_reference_middle is computed
        as the Y coordinate of the middle spectrum trail. The same value
        is used for all the available spectrum trails.
    list_spectrails: list of numpy.polynomial.Polynomial instances
        List of spectrum trails defined (lower, middle and upper).
    list_frontiers: list of numpy.polynomial.Polynomial instances
        List of spectrum trails defining the slitlet frontiers (lower
        and upper).
    y0_reference_lower: float
        Y coordinate corresponding to the lower spectrum trail computed
        at x0_reference. This value is employed as the Y coordinate of
        the lower spectrum trail of the rectified slitlet.
    y0_reference_middle: float
        Y coordinate corresponding to the middle spectrum trail computed
        at x0_reference. This value is employed as the Y coordinate of
        the middle spectrum trail of the rectified slitlet.
    y0_reference_upper: float
        Y coordinate corresponding to the upper spectrum trail computed
        at x0_reference. This value is employed as the Y coordinate of
        the upper spectrum trail of the rectified slitlet.
    y0_frontier_lower: float
        Y coordinate corresponding to the lower frontier computed at
        x0_reference.
    y0_frontier_upper: float
        Y coordinate corresponding to the upper frontier computed at
        x0_reference.
    y0_frontier_lower_expected: float
        Expected Y coordinate corresponding to the lower frontier
        computed at x0_reference in the rectified image.
    y0_frontier_upper_expected: float
        Expected Y coordinate corresponding to the upper frontier
        computed at x0_reference in the rectified image.
    corr_yrect_a : float
        Intercept of the relation y_expected = a + b * y_measured
        that transforms the measured ordinate into the expected ordinate
        of the rectified image.
    corr_yrect_b : float
        Slope of the relation y_expected = a + b * y_measured
        that transforms the measured ordinate into the expected ordinate
        of the rectified image.
    min_row_rectified : int
        Minimum useful row (starting from zero) of the rectified slitlet.
    max_row_rectified : int
        Maximum useful row (starting from zero) of the rectified slitlet.
    iminslt : int
        Minimum useful scan in output full 2d rectified image.
    imaxslt : int
        Maximum useful scan in output full 2d rectified image.
    jminslt : int
        Minimum useful channel in output full 2d rectified image.
    jmaxslt : int
        Maximum useful channel in output full 2d rectified image.
    ttd_order : int or None
        Polynomial order corresponding to the rectification
        transformation.
    ttd_aij : numpy array
        Polynomial coefficents corresponding to the direct
        rectification transformation coefficients a_ij.
    ttd_bij : numpy array
        Polynomial coefficents corresponding to the direct
        rectification transformation coefficients b_ij.
    tti_aij : numpy array
        Polynomial coefficents corresponding to the inverse
        rectification transformation coefficients a_ij.
    tti_bij : numpy array
        Polynomial coefficents corresponding to the inverse
        rectification transformation coefficients b_ij.
    wpoly : Polynomial instance
        Wavelength calibration polynomial, providing the
        wavelength as a function of pixel number (running from 1 to
        NAXIS1).
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.


    """

    def __init__(self, islitlet, rectwv_coeff, debugplot):
        # slitlet number
        self.islitlet = islitlet

        tmpcontent = rectwv_coeff.contents[islitlet-1]

        # csu configuration
        self.csu_bar_left = tmpcontent['csu_bar_left']
        self.csu_bar_right = tmpcontent['csu_bar_right']
        self.csu_bar_slit_center = tmpcontent['csu_bar_slit_center']
        self.csu_bar_slit_width = tmpcontent['csu_bar_slit_width']

        # horizontal and vertical bounding box
        self.bb_nc1_orig = tmpcontent['bb_nc1_orig']
        self.bb_nc2_orig = tmpcontent['bb_nc2_orig']
        self.bb_ns1_orig = tmpcontent['bb_ns1_orig']
        self.bb_ns2_orig = tmpcontent['bb_ns2_orig']

        # reference abscissa
        self.x0_reference = tmpcontent['x0_reference']

        # list of spectrum trails (lower, middle, and upper)
        self.list_spectrails = []
        for idum, cdum in zip(range(3), ['lower', 'middle', 'upper']):
            coeff = tmpcontent['spectrail']['poly_coef_' + cdum]
            self.list_spectrails.append(np.polynomial.Polynomial(coeff))

        # define reference ordinates using lower, middle and upper spectrails
        # evaluated at x0_reference
        self.y0_reference_lower = tmpcontent['y0_reference_lower']
        self.y0_reference_middle = tmpcontent['y0_reference_middle']
        self.y0_reference_upper = tmpcontent['y0_reference_upper']

        # list of frontiers (lower and upper)
        self.list_frontiers = []
        for idum, cdum in zip(range(2), ['lower', 'upper']):
            coeff = tmpcontent['frontier']['poly_coef_' + cdum]
            self.list_frontiers.append(np.polynomial.Polynomial(coeff))

        # define frontier ordinates at x0_reference
        self.y0_frontier_lower = tmpcontent['y0_frontier_lower']
        self.y0_frontier_upper = tmpcontent['y0_frontier_upper']

        # define expected frontier ordinates at x0_reference and
        # associated linear transformation
        self.y0_frontier_lower_expected = \
            tmpcontent['y0_frontier_lower_expected']
        self.y0_frontier_upper_expected = \
            tmpcontent['y0_frontier_upper_expected']
        self.corr_yrect_a = tmpcontent['corr_yrect_a']
        self.corr_yrect_b = tmpcontent['corr_yrect_b']
        self.min_row_rectified = tmpcontent['min_row_rectified']
        self.max_row_rectified = tmpcontent['max_row_rectified']

        # define useful scan region in output rectified image
        self.iminslt = (islitlet - 1) * EMIR_NPIXPERSLIT_RECTIFIED + 1
        self.imaxslt = islitlet * EMIR_NPIXPERSLIT_RECTIFIED

        # define useful channel region in output rectified image
        self.jminslt = 0
        self.jmaxslt = 0

        # Rectification coefficients
        self.ttd_aij = tmpcontent['ttd_aij']
        self.ttd_bij = tmpcontent['ttd_bij']
        self.tti_aij = tmpcontent['tti_aij']
        self.tti_bij = tmpcontent['tti_bij']
        # determine order from number of coefficients
        ncoef = len(self.ttd_aij)
        self.ttd_order = order_fmap(ncoef)

        # Wavelength calibration coefficients
        self.wpoly = tmpcontent['wpoly_coeff']

        # debugplot
        self.debugplot = debugplot

    def __repr__(self):
        """Define printable representation of a Slitlet2D instance."""

        # string with all the information
        output = "<Slilet2D instance>\n" + \
                 "- islitlet..................: " + \
                 str(self.islitlet) + "\n" + \
                 "- csu_bar_left..............: " + \
                 str(self.csu_bar_left) + "\n" + \
                 "- csu_bar_right.............: " + \
                 str(self.csu_bar_right) + "\n" + \
                 "- csu_bar_slit_center.......: " + \
                 str(self.csu_bar_slit_center) + "\n" + \
                 "- csu_bar_slit_width........: " + \
                 str(self.csu_bar_slit_width) + "\n" + \
                 "- x0_reference..............: " + \
                 str(self.x0_reference) + "\n" + \
                 "- y0_reference_lower........: " + \
                 str(self.y0_reference_lower) + "\n" + \
                 "- y0_reference_middle.......: " + \
                 str(self.y0_reference_middle) + "\n" + \
                 "- y0_reference_upper........: " + \
                 str(self.y0_reference_upper) + "\n" + \
                 "- y0_frontier_lower.........: " + \
                 str(self.y0_frontier_lower) + "\n" + \
                 "- y0_frontier_upper.........: " + \
                 str(self.y0_frontier_upper) + "\n" + \
                 "- y0_frontier_lower_expected: " + \
                 str(self.y0_frontier_lower_expected) + "\n" + \
                 "- y0_frontier_upper_expected: " + \
                 str(self.y0_frontier_upper_expected) + "\n" + \
                 "- corr_yrect_a................: " + \
                 str(self.corr_yrect_a) + "\n" + \
                 "- corr_yrect_b................: " + \
                 str(self.corr_yrect_b) + "\n" + \
                 "- min_row_rectified...........: " + \
                 str(self.min_row_rectified) + "\n" + \
                 "- max_row_rectified...........: " + \
                 str(self.max_row_rectified) + "\n" + \
                 "- iminslt.....................: " + \
                 str(self.iminslt) + "\n" + \
                 "- imaxslt.....................: " + \
                 str(self.imaxslt) + "\n" + \
                 "- jminslt.....................: " + \
                 str(self.jminslt) + "\n" + \
                 "- jmaxslt.....................: " + \
                 str(self.jmaxslt) + "\n" + \
                 "- bb_nc1_orig...............: " + \
                 str(self.bb_nc1_orig) + "\n" + \
                 "- bb_nc2_orig...............: " + \
                 str(self.bb_nc2_orig) + "\n" + \
                 "- bb_ns1_orig...............: " + \
                 str(self.bb_ns1_orig) + "\n" + \
                 "- bb_ns2_orig...............: " + \
                 str(self.bb_ns2_orig) + "\n" + \
                 "- lower spectrail...........:\n\t" + \
                 str(self.list_spectrails[0]) + "\n" + \
                 "- middle spectrail..........:\n\t" + \
                 str(self.list_spectrails[1]) + "\n" + \
                 "- upper spectrail...........:\n\t" + \
                 str(self.list_spectrails[2]) + "\n" + \
                 "- lower frontier............:\n\t" + \
                 str(self.list_frontiers[0]) + "\n" + \
                 "- upper frontier............:\n\t" + \
                 str(self.list_frontiers[1]) + "\n" + \
                 "- ttd_order.................: " + \
                 str(self.ttd_order) + "\n" + \
                 "- ttd_aij...................:\n\t" + \
                 str(self.ttd_aij) + "\n" + \
                 "- ttd_bij...................:\n\t" + \
                 str(self.ttd_bij) + "\n" + \
                 "- tti_aij...................:\n\t" + \
                 str(self.tti_aij) + "\n" + \
                 "- tti_bij...................:\n\t" + \
                 str(self.tti_bij) + "\n" + \
                 "- wpoly.....................:\n\t" + \
                 str(self.wpoly) + "\n" + \
                 "- debugplot.................: " + \
                 str(self.debugplot)

        return output

    def extract_slitlet2d(self, image_2k2k, subtitle=None):
        """Extract slitlet 2d image from image with original EMIR dimensions.

        Parameters
        ----------
        image_2k2k : numpy array
            Original image (dimensions EMIR_NAXIS1 * EMIR_NAXIS2)
        subtitle : string, optional
            Subtitle for plot.

        Returns
        -------
        slitlet2d : numpy array
            Image corresponding to the slitlet region defined by its
            bounding box.

        """

        # protections
        naxis2, naxis1 = image_2k2k.shape
        if naxis1 != EMIR_NAXIS1:
            raise ValueError('Unexpected naxis1')
        if naxis2 != EMIR_NAXIS2:
            raise ValueError('Unexpected naxis2')

        # extract slitlet region
        slitlet2d = image_2k2k[(self.bb_ns1_orig - 1):self.bb_ns2_orig,
                               (self.bb_nc1_orig - 1):self.bb_nc2_orig]

        # transform to float
        slitlet2d = slitlet2d.astype(np.float)

        # display slitlet2d with boundaries and middle spectrum trail
        if abs(self.debugplot) in [21, 22]:
            self.ximshow_unrectified(slitlet2d, subtitle)

        # return slitlet image
        return slitlet2d

    def rectify(self, slitlet2d, resampling, inverse=False, subtitle=None):
        """Rectify slitlet using computed transformation.

        Parameters
        ----------
        slitlet2d : numpy array
            Image containing the 2d slitlet image.
        resampling : int
            1: nearest neighbour, 2: flux preserving interpolation.
        inverse : bool
            If true, the inverse rectification transformation is
            employed.
        subtitle : string, optional
            Subtitle for plot.

        Returns
        -------
        slitlet2d_rect : numpy array
            Rectified slitlet image.

        """

        if resampling not in [1, 2]:
            raise ValueError("Unexpected resampling value=" + str(resampling))

        # check image dimension
        naxis2, naxis1 = slitlet2d.shape
        if naxis1 != self.bb_nc2_orig - self.bb_nc1_orig + 1:
            raise ValueError("Unexpected slitlet2d_rect naxis1")
        if naxis2 != self.bb_ns2_orig - self.bb_ns1_orig + 1:
            raise ValueError("Unexpected slitlet2d_rect naxis2")

        if inverse:
            aij = self.tti_aij
            bij = self.tti_bij
        else:
            aij = self.ttd_aij
            bij = self.ttd_bij

        # rectify image
        slitlet2d_rect = rectify2d(
            image2d=slitlet2d,
            aij=aij,
            bij=bij,
            resampling=resampling
        )

        if abs(self.debugplot % 10) != 0:
            if inverse:
                self.ximshow_unrectified(slitlet2d_rect, subtitle=subtitle)
            else:
                self.ximshow_rectified(slitlet2d_rect, subtitle=subtitle)

        return slitlet2d_rect

    def ximshow_unrectified(self, slitlet2d, subtitle=None):
        """Display unrectified image with spectrails and frontiers.

        Parameters
        ----------
        slitlet2d : numpy array
            Array containing the unrectified slitlet image.
        subtitle : string, optional
            Subtitle for plot.

        """

        title = "Slitlet#" + str(self.islitlet)
        if subtitle is not None:
            title += ' (' + subtitle + ')'
        ax = ximshow(slitlet2d, title=title,
                     first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                     show=False)
        xdum = np.linspace(1, EMIR_NAXIS1, num=EMIR_NAXIS1)
        ylower = self.list_spectrails[0](xdum)
        ax.plot(xdum, ylower, 'b-')
        ymiddle = self.list_spectrails[1](xdum)
        ax.plot(xdum, ymiddle, 'b--')
        yupper = self.list_spectrails[2](xdum)
        ax.plot(xdum, yupper, 'b-')
        ylower_frontier = self.list_frontiers[0](xdum)
        ax.plot(xdum, ylower_frontier, 'b:')
        yupper_frontier = self.list_frontiers[1](xdum)
        ax.plot(xdum, yupper_frontier, 'b:')
        if title is not None:
            ax.set_title(title)
        pause_debugplot(debugplot=self.debugplot, pltshow=True)

    def ximshow_rectified(self, slitlet2d_rect, subtitle=None):
        """Display rectified image with spectrails and frontiers.

        Parameters
        ----------
        slitlet2d_rect : numpy array
            Array containing the rectified slitlet image.
        subtitle : string, optional
            Subtitle for plot.

        """

        title = "Slitlet#" + str(self.islitlet)
        if subtitle is not None:
            title += ' (' + subtitle + ')'
        ax = ximshow(slitlet2d_rect, title=title,
                     first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                     show=False)
        # grid with fitted transformation: spectrum trails
        xx = np.arange(0, self.bb_nc2_orig - self.bb_nc1_orig + 1,
                       dtype=np.float)
        for spectrail in self.list_spectrails:
            yy0 = self.corr_yrect_a + \
                  self.corr_yrect_b * spectrail(self.x0_reference)
            yy = np.tile([yy0 - self.bb_ns1_orig], xx.size)
            ax.plot(xx + self.bb_nc1_orig, yy + self.bb_ns1_orig, "b")
        for spectrail in self.list_frontiers:
            yy0 = self.corr_yrect_a +\
                  self.corr_yrect_b * spectrail(self.x0_reference)
            yy = np.tile([yy0 - self.bb_ns1_orig], xx.size)
            ax.plot(xx + self.bb_nc1_orig, yy + self.bb_ns1_orig, "b:")
        # show plot
        pause_debugplot(self.debugplot, pltshow=True)
